# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'llm_settings.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QCheckBox, QComboBox,
    QFrame, QGridLayout, QGroupBox, QHBoxLayout,
    QHeaderView, QLabel, QLayout, QLineEdit,
    QPushButton, QScrollArea, QSizePolicy, QSpacerItem,
    QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget)

from airunner.components.application.gui.widgets.slider.slider_widget import SliderWidget

class Ui_llm_settings_widget(object):
    def setupUi(self, llm_settings_widget):
        if not llm_settings_widget.objectName():
            llm_settings_widget.setObjectName(u"llm_settings_widget")
        llm_settings_widget.resize(581, 742)
        self.gridLayout_4 = QGridLayout(llm_settings_widget)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.gridLayout_4.setHorizontalSpacing(0)
        self.gridLayout_4.setVerticalSpacing(10)
        self.gridLayout_4.setContentsMargins(10, 10, 10, 10)
        self.scrollArea = QScrollArea(llm_settings_widget)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setFrameShadow(QFrame.Shadow.Plain)
        self.scrollArea.setLineWidth(0)
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 559, 499))
        self.gridLayout = QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setHorizontalSpacing(0)
        self.gridLayout.setVerticalSpacing(10)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.prompt_template_container = QWidget(self.scrollAreaWidgetContents)
        self.prompt_template_container.setObjectName(u"prompt_template_container")
        self.gridLayout_17 = QGridLayout(self.prompt_template_container)
        self.gridLayout_17.setObjectName(u"gridLayout_17")
        self.gridLayout_17.setHorizontalSpacing(0)
        self.gridLayout_17.setVerticalSpacing(10)
        self.gridLayout_17.setContentsMargins(0, 0, 0, 0)

        self.gridLayout.addWidget(self.prompt_template_container, 0, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 263, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 2, 0, 1, 1)

        self.override_parameters = QGroupBox(self.scrollAreaWidgetContents)
        self.override_parameters.setObjectName(u"override_parameters")
        font = QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.override_parameters.setFont(font)
        self.override_parameters.setCheckable(True)
        self.override_parameters.setChecked(True)
        self.gridLayout_12 = QGridLayout(self.override_parameters)
        self.gridLayout_12.setSpacing(10)
        self.gridLayout_12.setObjectName(u"gridLayout_12")
        self.gridLayout_12.setContentsMargins(10, 10, 10, 10)
        self.widget_19 = QWidget(self.override_parameters)
        self.widget_19.setObjectName(u"widget_19")
        self.gridLayout_2 = QGridLayout(self.widget_19)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setHorizontalSpacing(10)
        self.gridLayout_2.setVerticalSpacing(0)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.seed = QLineEdit(self.widget_19)
        self.seed.setObjectName(u"seed")
        self.seed.setMinimumSize(QSize(0, 20))

        self.gridLayout_2.addWidget(self.seed, 0, 0, 1, 1)

        self.random_seed = QCheckBox(self.widget_19)
        self.random_seed.setObjectName(u"random_seed")
        self.random_seed.setMinimumSize(QSize(0, 20))

        self.gridLayout_2.addWidget(self.random_seed, 0, 1, 1, 1)


        self.gridLayout_12.addWidget(self.widget_19, 6, 0, 1, 1)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setSpacing(10)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.length_penalty = SliderWidget(self.override_parameters)
        self.length_penalty.setObjectName(u"length_penalty")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.length_penalty.sizePolicy().hasHeightForWidth())
        self.length_penalty.setSizePolicy(sizePolicy)
        self.length_penalty.setMinimumSize(QSize(0, 0))
        self.length_penalty.setProperty(u"slider_minimum", -100)
        self.length_penalty.setProperty(u"slider_maximum", 1000)
        self.length_penalty.setProperty(u"spinbox_minimum", 0.000000000000000)
        self.length_penalty.setProperty(u"spinbox_maximum", 1.000000000000000)
        self.length_penalty.setProperty(u"display_as_float", True)
        self.length_penalty.setProperty(u"slider_single_step", 1)
        self.length_penalty.setProperty(u"slider_page_step", 10)
        self.length_penalty.setProperty(u"spinbox_single_step", 0.010000000000000)
        self.length_penalty.setProperty(u"spinbox_page_step", 0.100000000000000)

        self.horizontalLayout_3.addWidget(self.length_penalty)

        self.num_beams = SliderWidget(self.override_parameters)
        self.num_beams.setObjectName(u"num_beams")
        sizePolicy.setHeightForWidth(self.num_beams.sizePolicy().hasHeightForWidth())
        self.num_beams.setSizePolicy(sizePolicy)
        self.num_beams.setMinimumSize(QSize(0, 0))
        self.num_beams.setProperty(u"slider_minimum", 1)
        self.num_beams.setProperty(u"slider_maximum", 100)
        self.num_beams.setProperty(u"spinbox_minimum", 0.000000000000000)
        self.num_beams.setProperty(u"spinbox_maximum", 100.000000000000000)
        self.num_beams.setProperty(u"display_as_float", False)
        self.num_beams.setProperty(u"slider_single_step", 1)
        self.num_beams.setProperty(u"slider_page_step", 10)
        self.num_beams.setProperty(u"spinbox_single_step", 0.010000000000000)
        self.num_beams.setProperty(u"spinbox_page_step", 0.100000000000000)

        self.horizontalLayout_3.addWidget(self.num_beams)


        self.gridLayout_12.addLayout(self.horizontalLayout_3, 3, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(10)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.top_p = SliderWidget(self.override_parameters)
        self.top_p.setObjectName(u"top_p")
        sizePolicy.setHeightForWidth(self.top_p.sizePolicy().hasHeightForWidth())
        self.top_p.setSizePolicy(sizePolicy)
        self.top_p.setMinimumSize(QSize(0, 0))
        self.top_p.setProperty(u"slider_minimum", 1)
        self.top_p.setProperty(u"slider_maximum", 1000)
        self.top_p.setProperty(u"spinbox_minimum", 0.000000000000000)
        self.top_p.setProperty(u"spinbox_maximum", 1.000000000000000)
        self.top_p.setProperty(u"display_as_float", True)
        self.top_p.setProperty(u"slider_single_step", 1)
        self.top_p.setProperty(u"slider_page_step", 10)
        self.top_p.setProperty(u"spinbox_single_step", 0.010000000000000)
        self.top_p.setProperty(u"spinbox_page_step", 0.100000000000000)

        self.horizontalLayout.addWidget(self.top_p)

        self.max_new_tokens = SliderWidget(self.override_parameters)
        self.max_new_tokens.setObjectName(u"max_new_tokens")
        sizePolicy.setHeightForWidth(self.max_new_tokens.sizePolicy().hasHeightForWidth())
        self.max_new_tokens.setSizePolicy(sizePolicy)
        self.max_new_tokens.setMinimumSize(QSize(0, 0))
        self.max_new_tokens.setProperty(u"slider_minimum", 1)
        self.max_new_tokens.setProperty(u"slider_maximum", 2556)
        self.max_new_tokens.setProperty(u"spinbox_minimum", 1.000000000000000)
        self.max_new_tokens.setProperty(u"spinbox_maximum", 2556.000000000000000)
        self.max_new_tokens.setProperty(u"display_as_float", False)
        self.max_new_tokens.setProperty(u"slider_single_step", 1)
        self.max_new_tokens.setProperty(u"slider_page_step", 2556)
        self.max_new_tokens.setProperty(u"spinbox_single_step", 1)
        self.max_new_tokens.setProperty(u"spinbox_page_step", 2556)

        self.horizontalLayout.addWidget(self.max_new_tokens)


        self.gridLayout_12.addLayout(self.horizontalLayout, 0, 0, 1, 1)

        self.line = QFrame(self.override_parameters)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout_12.addWidget(self.line, 9, 0, 1, 1)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setSpacing(10)
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.sequences = SliderWidget(self.override_parameters)
        self.sequences.setObjectName(u"sequences")
        sizePolicy.setHeightForWidth(self.sequences.sizePolicy().hasHeightForWidth())
        self.sequences.setSizePolicy(sizePolicy)
        self.sequences.setMinimumSize(QSize(0, 0))
        self.sequences.setProperty(u"slider_minimum", 1)
        self.sequences.setProperty(u"slider_maximum", 100)
        self.sequences.setProperty(u"spinbox_minimum", 0.000000000000000)
        self.sequences.setProperty(u"spinbox_maximum", 100.000000000000000)
        self.sequences.setProperty(u"display_as_float", False)
        self.sequences.setProperty(u"slider_single_step", 1)
        self.sequences.setProperty(u"slider_page_step", 10)
        self.sequences.setProperty(u"spinbox_single_step", 0.010000000000000)
        self.sequences.setProperty(u"spinbox_page_step", 0.100000000000000)

        self.horizontalLayout_6.addWidget(self.sequences)

        self.top_k = SliderWidget(self.override_parameters)
        self.top_k.setObjectName(u"top_k")
        sizePolicy.setHeightForWidth(self.top_k.sizePolicy().hasHeightForWidth())
        self.top_k.setSizePolicy(sizePolicy)
        self.top_k.setMinimumSize(QSize(0, 0))
        self.top_k.setProperty(u"slider_minimum", 0)
        self.top_k.setProperty(u"slider_maximum", 256)
        self.top_k.setProperty(u"spinbox_minimum", 0.000000000000000)
        self.top_k.setProperty(u"spinbox_maximum", 256.000000000000000)
        self.top_k.setProperty(u"display_as_float", False)
        self.top_k.setProperty(u"slider_single_step", 1)
        self.top_k.setProperty(u"slider_page_step", 10)
        self.top_k.setProperty(u"spinbox_single_step", 1)
        self.top_k.setProperty(u"spinbox_page_step", 10)

        self.horizontalLayout_6.addWidget(self.top_k)


        self.gridLayout_12.addLayout(self.horizontalLayout_6, 5, 0, 1, 1)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setSpacing(10)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.ngram_size = SliderWidget(self.override_parameters)
        self.ngram_size.setObjectName(u"ngram_size")
        sizePolicy.setHeightForWidth(self.ngram_size.sizePolicy().hasHeightForWidth())
        self.ngram_size.setSizePolicy(sizePolicy)
        self.ngram_size.setMinimumSize(QSize(0, 0))
        self.ngram_size.setProperty(u"slider_minimum", 0)
        self.ngram_size.setProperty(u"slider_maximum", 20)
        self.ngram_size.setProperty(u"spinbox_minimum", 0.000000000000000)
        self.ngram_size.setProperty(u"spinbox_maximum", 20.000000000000000)
        self.ngram_size.setProperty(u"display_as_float", False)
        self.ngram_size.setProperty(u"slider_single_step", 1)
        self.ngram_size.setProperty(u"slider_page_step", 1)
        self.ngram_size.setProperty(u"spinbox_single_step", 1.000000000000000)
        self.ngram_size.setProperty(u"spinbox_page_step", 1.000000000000000)

        self.horizontalLayout_5.addWidget(self.ngram_size)

        self.temperature = SliderWidget(self.override_parameters)
        self.temperature.setObjectName(u"temperature")
        sizePolicy.setHeightForWidth(self.temperature.sizePolicy().hasHeightForWidth())
        self.temperature.setSizePolicy(sizePolicy)
        self.temperature.setMinimumSize(QSize(0, 0))
        self.temperature.setProperty(u"slider_minimum", 1)
        self.temperature.setProperty(u"slider_maximum", 10000)
        self.temperature.setProperty(u"spinbox_minimum", 0.000100000000000)
        self.temperature.setProperty(u"spinbox_maximum", 1.000000000000000)
        self.temperature.setProperty(u"display_as_float", True)
        self.temperature.setProperty(u"slider_single_step", 1)
        self.temperature.setProperty(u"slider_page_step", 10)
        self.temperature.setProperty(u"spinbox_single_step", 0.010000000000000)
        self.temperature.setProperty(u"spinbox_page_step", 0.100000000000000)

        self.horizontalLayout_5.addWidget(self.temperature)


        self.gridLayout_12.addLayout(self.horizontalLayout_5, 4, 0, 1, 1)

        self.gridLayout_3 = QGridLayout()
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setHorizontalSpacing(0)
        self.gridLayout_3.setVerticalSpacing(10)
        self.early_stopping = QCheckBox(self.override_parameters)
        self.early_stopping.setObjectName(u"early_stopping")

        self.gridLayout_3.addWidget(self.early_stopping, 0, 0, 1, 1)

        self.use_cache = QCheckBox(self.override_parameters)
        self.use_cache.setObjectName(u"use_cache")

        self.gridLayout_3.addWidget(self.use_cache, 0, 2, 1, 1)

        self.do_sample = QCheckBox(self.override_parameters)
        self.do_sample.setObjectName(u"do_sample")

        self.gridLayout_3.addWidget(self.do_sample, 0, 1, 1, 1)


        self.gridLayout_12.addLayout(self.gridLayout_3, 8, 0, 1, 1)

        self.gridLayout_mode_routing = QGridLayout()
        self.gridLayout_mode_routing.setObjectName(u"gridLayout_mode_routing")
        self.gridLayout_mode_routing.setHorizontalSpacing(0)
        self.gridLayout_mode_routing.setVerticalSpacing(10)
        self.use_mode_routing = QCheckBox(self.override_parameters)
        self.use_mode_routing.setObjectName(u"use_mode_routing")

        self.gridLayout_mode_routing.addWidget(self.use_mode_routing, 0, 0, 1, 1)

        self.enable_trajectory_logging = QCheckBox(self.override_parameters)
        self.enable_trajectory_logging.setObjectName(u"enable_trajectory_logging")

        self.gridLayout_mode_routing.addWidget(self.enable_trajectory_logging, 0, 1, 1, 1)

        self.horizontalLayout_mode_override = QHBoxLayout()
        self.horizontalLayout_mode_override.setObjectName(u"horizontalLayout_mode_override")
        self.label_mode_override = QLabel(self.override_parameters)
        self.label_mode_override.setObjectName(u"label_mode_override")

        self.horizontalLayout_mode_override.addWidget(self.label_mode_override)

        self.mode_override = QComboBox(self.override_parameters)
        self.mode_override.addItem("")
        self.mode_override.addItem("")
        self.mode_override.addItem("")
        self.mode_override.addItem("")
        self.mode_override.addItem("")
        self.mode_override.addItem("")
        self.mode_override.setObjectName(u"mode_override")

        self.horizontalLayout_mode_override.addWidget(self.mode_override)

        self.horizontalSpacer_mode = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_mode_override.addItem(self.horizontalSpacer_mode)


        self.gridLayout_mode_routing.addLayout(self.horizontalLayout_mode_override, 1, 0, 1, 2)


        self.gridLayout_12.addLayout(self.gridLayout_mode_routing, 9, 0, 1, 1)

        self.pushButton = QPushButton(self.override_parameters)
        self.pushButton.setObjectName(u"pushButton")
        font1 = QFont()
        font1.setPointSize(9)
        font1.setBold(False)
        self.pushButton.setFont(font1)

        self.gridLayout_12.addWidget(self.pushButton, 11, 0, 1, 1)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setSpacing(10)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.repetition_penalty = SliderWidget(self.override_parameters)
        self.repetition_penalty.setObjectName(u"repetition_penalty")
        sizePolicy.setHeightForWidth(self.repetition_penalty.sizePolicy().hasHeightForWidth())
        self.repetition_penalty.setSizePolicy(sizePolicy)
        self.repetition_penalty.setMinimumSize(QSize(0, 0))
        self.repetition_penalty.setProperty(u"slider_minimum", 1)
        self.repetition_penalty.setProperty(u"slider_maximum", 10000)
        self.repetition_penalty.setProperty(u"spinbox_minimum", 0.010000000000000)
        self.repetition_penalty.setProperty(u"spinbox_maximum", 100.000000000000000)
        self.repetition_penalty.setProperty(u"display_as_float", True)
        self.repetition_penalty.setProperty(u"slider_single_step", 0)
        self.repetition_penalty.setProperty(u"slider_page_step", 1)
        self.repetition_penalty.setProperty(u"spinbox_single_step", 1.000000000000000)
        self.repetition_penalty.setProperty(u"spinbox_page_step", 10.000000000000000)

        self.horizontalLayout_2.addWidget(self.repetition_penalty)

        self.min_length = SliderWidget(self.override_parameters)
        self.min_length.setObjectName(u"min_length")
        sizePolicy.setHeightForWidth(self.min_length.sizePolicy().hasHeightForWidth())
        self.min_length.setSizePolicy(sizePolicy)
        self.min_length.setMinimumSize(QSize(0, 0))
        self.min_length.setProperty(u"slider_minimum", 1)
        self.min_length.setProperty(u"slider_maximum", 2556)
        self.min_length.setProperty(u"spinbox_minimum", 1.000000000000000)
        self.min_length.setProperty(u"spinbox_maximum", 2556.000000000000000)
        self.min_length.setProperty(u"display_as_float", False)
        self.min_length.setProperty(u"slider_single_step", 1)
        self.min_length.setProperty(u"slider_page_step", 2556)
        self.min_length.setProperty(u"spinbox_single_step", 1)
        self.min_length.setProperty(u"spinbox_page_step", 2556)

        self.horizontalLayout_2.addWidget(self.min_length)


        self.gridLayout_12.addLayout(self.horizontalLayout_2, 2, 0, 1, 1)


        self.gridLayout.addWidget(self.override_parameters, 1, 0, 1, 1)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.gridLayout_4.addWidget(self.scrollArea, 4, 0, 1, 1)

        self.label_2 = QLabel(llm_settings_widget)
        self.label_2.setObjectName(u"label_2")
        font2 = QFont()
        font2.setPointSize(11)
        font2.setBold(True)
        self.label_2.setFont(font2)

        self.gridLayout_4.addWidget(self.label_2, 0, 0, 1, 1)

        self.line_2 = QFrame(llm_settings_widget)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.HLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout_4.addWidget(self.line_2, 1, 0, 1, 1)

        self.groupBox = QGroupBox(llm_settings_widget)
        self.groupBox.setObjectName(u"groupBox")
        font3 = QFont()
        font3.setPointSize(10)
        font3.setBold(True)
        self.groupBox.setFont(font3)
        self.verticalLayout = QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(9, 9, 9, 9)
        self.model_service = QComboBox(self.groupBox)
        self.model_service.setObjectName(u"model_service")

        self.verticalLayout.addWidget(self.model_service)


        self.gridLayout_4.addWidget(self.groupBox, 2, 0, 1, 1)

        self.model_selection_group = QGroupBox(llm_settings_widget)
        self.model_selection_group.setObjectName(u"model_selection_group")
        self.model_selection_group.setFont(font3)
        self.verticalLayout_2 = QVBoxLayout(self.model_selection_group)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(9, 9, 9, 9)
        self.model_dropdown = QComboBox(self.model_selection_group)
        self.model_dropdown.setObjectName(u"model_dropdown")
        self.model_dropdown.setEditable(True)

        self.verticalLayout_2.addWidget(self.model_dropdown)

        self.model_path = QLineEdit(self.model_selection_group)
        self.model_path.setObjectName(u"model_path")

        self.verticalLayout_2.addWidget(self.model_path)

        self.quantization_layout = QVBoxLayout()
        self.quantization_layout.setObjectName(u"quantization_layout")
        self.quantization_label = QLabel(self.model_selection_group)
        self.quantization_label.setObjectName(u"quantization_label")
        self.quantization_label.setFont(font1)

        self.quantization_layout.addWidget(self.quantization_label)

        self.quantization_dropdown = QComboBox(self.model_selection_group)
        self.quantization_dropdown.addItem("")
        self.quantization_dropdown.addItem("")
        self.quantization_dropdown.addItem("")
        self.quantization_dropdown.addItem("")
        self.quantization_dropdown.setObjectName(u"quantization_dropdown")
        self.quantization_dropdown.setEnabled(False)
        self.quantization_dropdown.setFont(font1)

        self.quantization_layout.addWidget(self.quantization_dropdown)


        self.verticalLayout_2.addLayout(self.quantization_layout)

        self.model_info_label = QLabel(self.model_selection_group)
        self.model_info_label.setObjectName(u"model_info_label")
        font4 = QFont()
        font4.setPointSize(9)
        font4.setBold(False)
        font4.setItalic(True)
        self.model_info_label.setFont(font4)
        self.model_info_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.model_info_label)

        self.download_model_button = QPushButton(self.model_selection_group)
        self.download_model_button.setObjectName(u"download_model_button")
        self.download_model_button.setEnabled(False)

        self.verticalLayout_2.addWidget(self.download_model_button)

        self.start_quantize_button = QPushButton(self.model_selection_group)
        self.start_quantize_button.setObjectName(u"start_quantize_button")
        self.start_quantize_button.setEnabled(False)

        self.verticalLayout_2.addWidget(self.start_quantize_button)

        self.delete_buttons_layout = QHBoxLayout()
        self.delete_buttons_layout.setObjectName(u"delete_buttons_layout")
        self.delete_safetensors_button = QPushButton(self.model_selection_group)
        self.delete_safetensors_button.setObjectName(u"delete_safetensors_button")
        self.delete_safetensors_button.setEnabled(False)

        self.delete_buttons_layout.addWidget(self.delete_safetensors_button)

        self.delete_quantized_button = QPushButton(self.model_selection_group)
        self.delete_quantized_button.setObjectName(u"delete_quantized_button")
        self.delete_quantized_button.setEnabled(False)

        self.delete_buttons_layout.addWidget(self.delete_quantized_button)


        self.verticalLayout_2.addLayout(self.delete_buttons_layout)


        self.gridLayout_4.addWidget(self.model_selection_group, 3, 0, 1, 1)

        self.adapters_group = QGroupBox(llm_settings_widget)
        self.adapters_group.setObjectName(u"adapters_group")
        self.adapters_group.setFont(font3)
        self.verticalLayout_3 = QVBoxLayout(self.adapters_group)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(9, 9, 9, 9)
        self.adapters_table = QTableWidget(self.adapters_group)
        if (self.adapters_table.columnCount() < 3):
            self.adapters_table.setColumnCount(3)
        __qtablewidgetitem = QTableWidgetItem()
        self.adapters_table.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.adapters_table.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.adapters_table.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        self.adapters_table.setObjectName(u"adapters_table")
        self.adapters_table.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.adapters_table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.adapters_table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)

        self.verticalLayout_3.addWidget(self.adapters_table)

        self.refresh_adapters_button = QPushButton(self.adapters_group)
        self.refresh_adapters_button.setObjectName(u"refresh_adapters_button")

        self.verticalLayout_3.addWidget(self.refresh_adapters_button)


        self.gridLayout_4.addWidget(self.adapters_group, 5, 0, 1, 1)

        QWidget.setTabOrder(self.seed, self.random_seed)

        self.retranslateUi(llm_settings_widget)
        self.override_parameters.toggled.connect(llm_settings_widget.override_parameters_toggled)
        self.early_stopping.toggled.connect(llm_settings_widget.early_stopping_toggled)
        self.pushButton.clicked.connect(llm_settings_widget.reset_settings_to_default_clicked)
        self.do_sample.toggled.connect(llm_settings_widget.do_sample_toggled)
        self.use_cache.clicked["bool"].connect(llm_settings_widget.toggle_use_cache)
        self.quantization_dropdown.currentIndexChanged.connect(llm_settings_widget.on_quantization_changed)

        QMetaObject.connectSlotsByName(llm_settings_widget)
    # setupUi

    def retranslateUi(self, llm_settings_widget):
        llm_settings_widget.setWindowTitle(QCoreApplication.translate("llm_settings_widget", u"Form", None))
        self.override_parameters.setTitle(QCoreApplication.translate("llm_settings_widget", u"Override Prameters", None))
        self.random_seed.setText(QCoreApplication.translate("llm_settings_widget", u"Random seed", None))
        self.length_penalty.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.length_penalty", None))
        self.length_penalty.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Length Penalty", None))
        self.num_beams.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.num_beams", None))
        self.num_beams.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Num Beams", None))
        self.top_p.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.top_p", None))
        self.top_p.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Top P", None))
        self.max_new_tokens.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.max_new_tokens", None))
        self.max_new_tokens.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Max New Tokens", None))
        self.sequences.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.sequences", None))
        self.sequences.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Sequences", None))
        self.top_k.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.top_k", None))
        self.top_k.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Top K", None))
        self.ngram_size.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.ngram_size", None))
        self.ngram_size.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Ngram Size", None))
        self.temperature.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.temperature", None))
        self.temperature.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Temperature", None))
        self.early_stopping.setText(QCoreApplication.translate("llm_settings_widget", u"Early stopping", None))
        self.use_cache.setText(QCoreApplication.translate("llm_settings_widget", u"Use Cache", None))
        self.do_sample.setText(QCoreApplication.translate("llm_settings_widget", u"Do sample", None))
        self.use_mode_routing.setText(QCoreApplication.translate("llm_settings_widget", u"Enable Mode-Based Routing", None))
#if QT_CONFIG(tooltip)
        self.use_mode_routing.setToolTip(QCoreApplication.translate("llm_settings_widget", u"Use specialized agents for author, code, research, and QA tasks", None))
#endif // QT_CONFIG(tooltip)
        self.enable_trajectory_logging.setText(QCoreApplication.translate("llm_settings_widget", u"Log Trajectories (Debug)", None))
#if QT_CONFIG(tooltip)
        self.enable_trajectory_logging.setToolTip(QCoreApplication.translate("llm_settings_widget", u"Enable detailed logging of agent trajectories for debugging", None))
#endif // QT_CONFIG(tooltip)
        self.label_mode_override.setText(QCoreApplication.translate("llm_settings_widget", u"Force Mode:", None))
        self.mode_override.setItemText(0, QCoreApplication.translate("llm_settings_widget", u"Auto (Default)", None))
        self.mode_override.setItemText(1, QCoreApplication.translate("llm_settings_widget", u"Author", None))
        self.mode_override.setItemText(2, QCoreApplication.translate("llm_settings_widget", u"Code", None))
        self.mode_override.setItemText(3, QCoreApplication.translate("llm_settings_widget", u"Research", None))
        self.mode_override.setItemText(4, QCoreApplication.translate("llm_settings_widget", u"QA", None))
        self.mode_override.setItemText(5, QCoreApplication.translate("llm_settings_widget", u"General", None))

#if QT_CONFIG(tooltip)
        self.mode_override.setToolTip(QCoreApplication.translate("llm_settings_widget", u"Override automatic mode detection (leave empty for automatic)", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton.setText(QCoreApplication.translate("llm_settings_widget", u"Reset Settings to Default", None))
        self.repetition_penalty.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.repetition_penalty", None))
        self.repetition_penalty.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Repetition Penalty", None))
        self.min_length.setProperty(u"settings_property", QCoreApplication.translate("llm_settings_widget", u"llm_generator_settings.min_length", None))
        self.min_length.setProperty(u"label_text", QCoreApplication.translate("llm_settings_widget", u"Min Length", None))
        self.label_2.setText(QCoreApplication.translate("llm_settings_widget", u"LLM Settings", None))
        self.groupBox.setTitle(QCoreApplication.translate("llm_settings_widget", u"Provider", None))
        self.model_selection_group.setTitle(QCoreApplication.translate("llm_settings_widget", u"Model", None))
        self.model_path.setPlaceholderText(QCoreApplication.translate("llm_settings_widget", u"Model path or identifier", None))
        self.quantization_label.setText(QCoreApplication.translate("llm_settings_widget", u"Quantization:", None))
        self.quantization_dropdown.setItemText(0, QCoreApplication.translate("llm_settings_widget", u"2-bit (50% VRAM, -15% quality)", None))
        self.quantization_dropdown.setItemText(1, QCoreApplication.translate("llm_settings_widget", u"4-bit (balanced, recommended)", None))
        self.quantization_dropdown.setItemText(2, QCoreApplication.translate("llm_settings_widget", u"8-bit (200% VRAM, best quality)", None))
        self.quantization_dropdown.setItemText(3, QCoreApplication.translate("llm_settings_widget", u"GGUF Q4_K_M (smallest, fastest)", None))

        self.model_info_label.setText(QCoreApplication.translate("llm_settings_widget", u"VRAM: -- GB | Context: -- tokens | Tool Calling: --", None))
        self.download_model_button.setText(QCoreApplication.translate("llm_settings_widget", u"Download Model from HuggingFace", None))
        self.start_quantize_button.setText(QCoreApplication.translate("llm_settings_widget", u"Start Manual Quantization", None))
#if QT_CONFIG(tooltip)
        self.start_quantize_button.setToolTip(QCoreApplication.translate("llm_settings_widget", u"Manually quantize model to disk (creates separate files). Automatic quantization during loading is recommended instead.", None))
#endif // QT_CONFIG(tooltip)
        self.delete_safetensors_button.setText(QCoreApplication.translate("llm_settings_widget", u"Delete Original Files", None))
        self.delete_quantized_button.setText(QCoreApplication.translate("llm_settings_widget", u"Delete Quantized", None))
        self.adapters_group.setTitle(QCoreApplication.translate("llm_settings_widget", u"LoRA Adapters", None))
        ___qtablewidgetitem = self.adapters_table.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("llm_settings_widget", u"Enabled", None));
        ___qtablewidgetitem1 = self.adapters_table.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("llm_settings_widget", u"Adapter Name", None));
        ___qtablewidgetitem2 = self.adapters_table.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("llm_settings_widget", u"Path", None));
        self.refresh_adapters_button.setText(QCoreApplication.translate("llm_settings_widget", u"Refresh Adapters", None))
    # retranslateUi




from PySide6.QtCore import Slot
from PySide6.QtWidgets import QFileDialog

from airunner.components.application.gui.widgets.base_widget import BaseWidget
from airunner.components.tts.gui.widgets.templates.open_voice_preferences_ui import (
    Ui_open_voice_preferences,
)
from airunner.components.tts.data.models.openvoice_settings import OpenVoiceSettings
from airunner.enums import AvailableLanguage


class OpenVoicePreferencesWidget(BaseWidget):
    widget_class_ = Ui_open_voice_preferences

    def __init__(self, id: int, *args, **kwargs):
        self._id: int = id
        self._item: OpenVoiceSettings = OpenVoiceSettings.objects.get(self._id)
        if not self._item:
            self._item = OpenVoiceSettings.objects.create()
        super().__init__(*args, **kwargs)
        self.ui.voice_sample_path.setText(self._item.reference_speaker_path)

    @Slot()
    def on_browse_voice_sample_path_button_clicked(self):
        """Open a file dialog to select a voice sample path."""
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            self.tr("Select Voice Sample Path"),
            "",
            self.tr("Audio Files (*.wav *.mp3 *.ogg);;All Files (*)"),
        )
        if file_path:
            self.ui.voice_sample_path.setText(file_path)
            open_voice_settings = OpenVoiceSettings.objects.get(self._id)
            if not open_voice_settings:
                return
            OpenVoiceSettings.objects.update(
                self._id, reference_speaker_path=file_path
            )

    @Slot(str)
    def on_voice_sample_path_textChanged(self, txt: str):
        open_voice_settings = OpenVoiceSettings.objects.get(self._id)
        if not open_voice_settings:
            return
        if open_voice_settings.reference_speaker_path != txt:
            OpenVoiceSettings.objects.update(
                self._id, reference_speaker_path=txt
            )

    def initialize_ui(self):
        # initialize comboboxes
        self.ui.language_combobox.addItems(
            [lang.value for lang in AvailableLanguage]
        )

        # Set the default language to the first item in the combobox
        self.ui.speed_slider.setProperty("table_item", self._item)

    def showEvent(self, event):
        super().showEvent(event)
        self.load_settings()

    def load_settings(self):
        """Load the OpenVoice settings into the widget."""
        settings = OpenVoiceSettings.objects.get(self._id)
        if not settings:
            return
        self.ui.language_combobox.setCurrentText(settings.language)

    @Slot(str)
    def language_changed(self, text):
        OpenVoiceSettings.objects.update(self._id, language=text)

    @Slot(int)
    def speed_changed(self, value):
        OpenVoiceSettings.objects.update(self._id, speed=value)



"""
Mood and emotional state tools.

Tools for tracking and updating the chatbot's emotional state.
"""

from typing import Annotated, Any

from airunner.components.llm.core.tool_registry import tool, ToolCategory
from airunner.enums import SignalCode


@tool(
    name="update_mood",
    category=ToolCategory.MOOD,
    description="Update the chatbot's emotional state based on conversation",
    return_direct=False,
    requires_api=True,
)
def update_mood(
    mood: Annotated[
        str,
        "A one-word emotion (e.g., happy, sad, excited, confused, neutral)",
    ],
    emoji: Annotated[
        str, "A single emoji representing the mood (e.g., 😊, 😢, 😡, 😐)"
    ] = "😐",
    api: Any = None,
) -> str:
    """Update the chatbot's emotional state.

    This should be called periodically during conversation (approximately every
    3 conversation turns) to reflect how the bot is feeling based on the interaction.

    Args:
        mood: A one-word emotion describing the current state
        emoji: A single emoji representing the mood
        api: API instance (injected automatically)


    Examples:
        update_mood("happy", "😊")
        update_mood("confused", "🤔")
        update_mood("excited", "🎉")
    """
    try:
        # Emit signal to update bot mood in the UI
        from airunner.utils.application.signal_mediator import SignalMediator

        mediator = SignalMediator()
        mediator.emit_signal(
            SignalCode.BOT_MOOD_UPDATED,
            {"mood": mood, "emoji": emoji},
        )
        return f"Mood updated to '{mood}' {emoji}"
    except Exception as e:
        return f"Error updating mood: {str(e)}"



"""
Unit tests for intelligent_crawl tool.

Tests basic functionality of the intelligent web crawler tool,
including parameter validation and error handling.
"""

import unittest
from unittest.mock import Mock, patch

from airunner.components.llm.tools.intelligent_crawl_tool import (
    intelligent_crawl,
)


class TestIntelligentCrawlTool(unittest.TestCase):
    """Test intelligent_crawl tool basic functionality."""

    def setUp(self):
        """Set up test fixtures."""
        self.mock_api = Mock()
        self.mock_api.llm = Mock()
        self.mock_api.llm.send_request = Mock()

    def test_no_api_provided(self):
        """Test error handling when API is not provided."""
        result = intelligent_crawl(
            start_url="https://example.com",
            research_goal="Test research goal",
            api=None,
        )

        # Should return error message
        self.assertIn("Error", result)
        self.assertIn("API not available", result)

    def test_parameter_clamping_max_pages(self):
        """Test that max_pages is clamped to valid range."""
        with patch(
            "airunner.components.llm.tools.intelligent_crawl_tool.CrawlerProcess"
        ) as mock_crawler:
            mock_process = Mock()
            mock_crawler.return_value = mock_process

            # Test value too high - should be clamped
            result = intelligent_crawl(
                start_url="https://example.com",
                research_goal="Test",
                max_pages=100,  # Should be clamped to 20
                api=self.mock_api,
            )

            # Should still work (not crash)
            self.assertIsInstance(result, str)

            # Test value too low - should be clamped
            result = intelligent_crawl(
                start_url="https://example.com",
                research_goal="Test",
                max_pages=0,  # Should be clamped to 1
                api=self.mock_api,
            )

            # Should still work (not crash)
            self.assertIsInstance(result, str)

    def test_parameter_clamping_max_depth(self):
        """Test that max_depth is clamped to valid range."""
        with patch(
            "airunner.components.llm.tools.intelligent_crawl_tool.CrawlerProcess"
        ) as mock_crawler:
            mock_process = Mock()
            mock_crawler.return_value = mock_process

            # Test value too high - should be clamped
            result = intelligent_crawl(
                start_url="https://example.com",
                research_goal="Test",
                max_depth=10,  # Should be clamped to 3
                api=self.mock_api,
            )

            # Should still work (not crash)
            self.assertIsInstance(result, str)

            # Test value too low - should be clamped
            result = intelligent_crawl(
                start_url="https://example.com",
                research_goal="Test",
                max_depth=0,  # Should be clamped to 1
                api=self.mock_api,
            )

            # Should still work (not crash)
            self.assertIsInstance(result, str)

    @patch(
        "airunner.components.llm.tools.intelligent_crawl_tool.CrawlerProcess"
    )
    def test_crawler_process_created(self, mock_crawler_process):
        """Test that crawler process is created."""
        mock_process = Mock()
        mock_crawler_process.return_value = mock_process

        result = intelligent_crawl(
            start_url="https://example.com",
            research_goal="Test research",
            max_pages=5,
            max_depth=2,
            api=self.mock_api,
        )

        # Verify CrawlerProcess was created
        mock_crawler_process.assert_called_once()

        # Verify process was started
        mock_process.start.assert_called_once()

        # Result should be a string
        self.assertIsInstance(result, str)

    @patch(
        "airunner.components.llm.tools.intelligent_crawl_tool.CrawlerProcess"
    )
    def test_returns_string_result(self, mock_crawler_process):
        """Test that function returns a string result."""
        mock_process = Mock()
        mock_crawler_process.return_value = mock_process

        result = intelligent_crawl(
            start_url="https://example.com",
            research_goal="Test research",
            api=self.mock_api,
        )

        # Result should be a string
        self.assertIsInstance(result, str)

    @patch(
        "airunner.components.llm.tools.intelligent_crawl_tool.CrawlerProcess"
    )
    def test_crawler_exception_handling(self, mock_crawler_process):
        """Test that exceptions during crawling are handled gracefully."""
        # Make CrawlerProcess raise an exception
        mock_crawler_process.side_effect = Exception("Crawler error")

        result = intelligent_crawl(
            start_url="https://example.com",
            research_goal="Test research",
            api=self.mock_api,
        )

        # Should return error message, not raise exception
        self.assertIsInstance(result, str)
        self.assertIn("Error", result)


if __name__ == "__main__":
    unittest.main()



"""Mixin for recentering logic in CustomGraphicsView.

This mixin handles grid and layer recentering operations.
"""

from PySide6.QtCore import QPointF


class RecenteringMixin:
    """Provides recentering functionality for graphics view.

    This mixin manages:
    - Grid recentering
    - Layer position recentering
    - Canvas offset reset

    Dependencies:
        - self.scene: CustomScene instance
        - self.logger: Logging instance
        - self.canvas_offset: QPointF canvas offset property
        - self._grid_compensation_offset: QPointF grid compensation
        - self.center_pos: QPointF center position
        - self.application_settings: Application settings
        - self.get_recentered_position(): Calculate centered position
        - self.update_active_grid_settings(): Update grid settings
        - self.save_canvas_offset(): Save offset to settings
        - self.api: API instance
        - self.update_active_grid_area_position(): Update grid position
        - self.recenter_layer_positions(): Recenter all layers
        - self.update_drawing_pad_settings(): Update drawing pad settings
        - self.updateImagePositions(): Update image positions
    """

    def on_recenter_grid_signal(self):
        """Handle recenter grid signal.

        Resets canvas offset and grid compensation, then recenters
        the active grid area and all layer images to viewport center.
        """
        # Reset offsets
        self.canvas_offset = QPointF(0, 0)
        self._grid_compensation_offset = QPointF(0, 0)

        if not self.scene:
            return

        # Calculate new center position for active grid area
        pos_x, pos_y = self.get_recentered_position(
            self.application_settings.working_width,
            self.application_settings.working_height,
        )
        self.center_pos = QPointF(pos_x, pos_y)

        # Update grid settings with new position
        self.update_active_grid_settings(
            pos_x=pos_x,
            pos_y=pos_y,
        )

        # Save the reset offset
        self.save_canvas_offset()

        # Notify API of grid info update
        self.api.art.canvas.update_grid_info(
            {
                "offset_x": self.canvas_offset_x,
                "offset_y": self.canvas_offset_y,
            }
        )

        # Update active grid area display position
        self.update_active_grid_area_position()

        # Recalculate and save new centered positions
        new_positions = {}

        # Handle old single-item system (DrawingPad item)
        if self.scene.item:
            self.logger.info(
                "[RECENTER] Processing old single-item system (DrawingPad)"
            )
            item_rect = self.scene.item.boundingRect()
            image_width = item_rect.width()
            image_height = item_rect.height()
            item_pos_x, item_pos_y = self.get_recentered_position(
                int(image_width), int(image_height)
            )

            # Save to database
            self.update_drawing_pad_settings(
                x_pos=item_pos_x, y_pos=item_pos_y
            )

            # Add to positions dict
            new_positions[self.scene.item] = QPointF(item_pos_x, item_pos_y)
            self.logger.info(
                f"[RECENTER] DrawingPad item: saved to DB and dict - position x={item_pos_x}, y={item_pos_y}"
            )

        # Handle new layer system
        layer_positions = self.recenter_layer_positions()
        new_positions.update(layer_positions)

        self.logger.info(
            f"[RECENTER] Total items to recenter: {len(new_positions)}"
        )

        # Clear caches AFTER saving to DB but BEFORE updating scene
        # This ensures scene.update_image_position() uses fresh DB values
        if hasattr(self.scene, "original_item_positions"):
            self.scene.original_item_positions = {}

        # Update all image positions with the new centered positions
        self.updateImagePositions(new_positions)
        
        # Store the new positions for future updates
        if hasattr(self.scene, "original_item_positions"):
            self.scene.original_item_positions.update(new_positions)

        self.logger.info("[RECENTER] Recenter grid signal processing complete")



"""Base class for specialized expert agents."""

from abc import ABC, abstractmethod
from typing import Dict, List, Any, Optional
from dataclasses import dataclass


@dataclass
class AgentCapability:
    """Describes a capability that an expert agent possesses.

    Attributes:
        name: Unique identifier for the capability
        description: Human-readable description
        keywords: Keywords that trigger this capability
        priority: Higher values indicate stronger match (0-10)
    """

    name: str
    description: str
    keywords: List[str]
    priority: int = 5


class ExpertAgent(ABC):
    """Abstract base class for specialized expert agents.

    Expert agents are specialized AI assistants that excel at specific
    tasks or domains (e.g., calendar management, code generation, research).
    They provide capabilities, evaluate task relevance, and execute tasks
    within their domain of expertise.
    """

    def __init__(self, name: str, description: str):
        """Initialize expert agent.

        Args:
            name: Unique identifier for the agent
            description: Human-readable description of agent's purpose
        """
        self.name = name
        self.description = description
        self._capabilities: List[AgentCapability] = []

    @property
    def capabilities(self) -> List[AgentCapability]:
        """Get list of agent capabilities.

        Returns:
            List of AgentCapability objects
        """
        return self._capabilities

    def register_capability(
        self,
        name: str,
        description: str,
        keywords: List[str],
        priority: int = 5,
    ) -> None:
        """Register a new capability for this agent.

        Args:
            name: Capability identifier
            description: Human-readable description
            keywords: Keywords that trigger this capability
            priority: Priority level (0-10)
        """
        capability = AgentCapability(
            name=name,
            description=description,
            keywords=keywords,
            priority=priority,
        )
        self._capabilities.append(capability)

    def evaluate_task(
        self, task: str, context: Optional[Dict] = None
    ) -> float:
        """Evaluate how well this agent can handle the given task.

        Analyzes the task and context to determine relevance score.
        Higher scores indicate better suitability for the task.

        Args:
            task: Task description
            context: Optional context dictionary

        Returns:
            Relevance score (0.0 to 1.0)
        """
        task_lower = task.lower()
        context = context or {}

        # Check for keyword matches across all capabilities
        max_score = 0.0
        for capability in self.capabilities:
            keyword_matches = sum(
                1
                for keyword in capability.keywords
                if keyword.lower() in task_lower
            )
            if keyword_matches > 0:
                # Score based on keyword matches and priority
                score = (keyword_matches / len(capability.keywords)) * (
                    capability.priority / 10.0
                )
                max_score = max(max_score, score)

        return min(max_score, 1.0)

    @abstractmethod
    async def execute_task(
        self, task: str, context: Optional[Dict] = None
    ) -> Dict[str, Any]:
        """Execute the given task.

        Args:
            task: Task description
            context: Optional context dictionary

        Returns:
            Dictionary containing:
                - result: Task result
                - success: Boolean indicating success
                - metadata: Optional metadata about execution
        """

    def get_available_tools(self) -> List[str]:
        """Get list of available tool names for this agent.

        Returns:
            List of tool names
        """
        return []

    def __repr__(self) -> str:
        """String representation of the agent.

        Returns:
            String representation
        """
        return f"<ExpertAgent: {self.name}>"



"""Long-Running Agent Harness - Main orchestrator.

The LongRunningHarness is the primary interface for running multi-session
projects. It manages the lifecycle of projects from initialization through
completion.

Key responsibilities:
1. Project lifecycle management
2. Session orchestration
3. Sub-agent coordination
4. State recovery and resumption
5. Progress monitoring

This builds on Anthropic's approach with several enhancements:
- Decision memory for learning from outcomes
- Specialized sub-agents for different task types
- Sophisticated state recovery
- Resource-aware execution
"""

from typing import Any, Optional, List, Dict, Callable
from pathlib import Path
import json

from airunner.components.llm.long_running.data.project_state import (
    ProjectState,
    ProjectFeature,
    ProjectStatus,
    FeatureStatus,
    FeatureCategory,
    DecisionOutcome,
)
from airunner.components.llm.long_running.project_manager import ProjectManager
from airunner.components.llm.long_running.initializer_agent import (
    InitializerAgent,
)
from airunner.components.llm.long_running.session_agent import SessionAgent
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


class LongRunningHarness:
    """Main orchestrator for long-running agent projects.

    The harness manages multi-session projects following Anthropic's
    two-phase approach (initializer + session agents) with extensions
    for decision memory, sub-agent delegation, and state recovery.

    Example:
        ```python
        harness = LongRunningHarness(
            chat_model=my_llm,
            tools=my_tools,
        )

        # Start a new project
        project_id = harness.create_project(
            name="My Chat App",
            description="Build a real-time chat application with rooms and DMs",
            working_directory="/home/user/projects/chat-app"
        )

        # Run sessions until complete (or hit max)
        result = harness.run_until_complete(
            project_id,
            max_sessions=50
        )

        # Or run one session at a time
        session_result = harness.run_session(project_id)
        ```

    Attributes:
        chat_model: LangChain chat model for agents
        tools: List of tools available to agents
        project_manager: Manages project persistence
        initializer: InitializerAgent instance
        session_agent: SessionAgent instance
        sub_agents: Dict of specialized sub-agents by category
    """

    def __init__(
        self,
        chat_model: Any,
        tools: Optional[List[Any]] = None,
        project_manager: Optional[ProjectManager] = None,
        sub_agents: Optional[Dict[str, Any]] = None,
        on_progress: Optional[Callable[[Dict], None]] = None,
    ):
        """Initialize the harness.

        Args:
            chat_model: LangChain chat model
            tools: List of tools for agents
            project_manager: Optional custom ProjectManager
            sub_agents: Optional dict of specialized sub-agents
            on_progress: Optional callback for progress updates
        """
        self._chat_model = chat_model
        self._tools = tools or []
        self._project_manager = project_manager or ProjectManager()
        self._sub_agents = sub_agents or {}
        self._on_progress = on_progress

        # Create agents
        self._initializer = InitializerAgent(
            chat_model=self._chat_model,
            project_manager=self._project_manager,
        )
        self._session_agent = SessionAgent(
            chat_model=self._chat_model,
            tools=self._tools,
            project_manager=self._project_manager,
            sub_agents=self._sub_agents,
        )

        logger.info("LongRunningHarness initialized")

    def register_sub_agent(
        self,
        category: str,
        agent: Any,
    ) -> None:
        """Register a specialized sub-agent for a feature category.

        Sub-agents handle specific types of work:
        - "code": Code writing, debugging, testing
        - "research": Information gathering, synthesis
        - "documentation": Writing docs, comments
        - "testing": Test creation, validation

        Args:
            category: Feature category this agent handles
            agent: Agent instance with invoke() method
        """
        self._sub_agents[category] = agent
        # Update session agent with new sub-agents
        self._session_agent = SessionAgent(
            chat_model=self._chat_model,
            tools=self._tools,
            project_manager=self._project_manager,
            sub_agents=self._sub_agents,
        )
        logger.info(f"Registered sub-agent for category: {category}")

    def create_project(
        self,
        name: str,
        description: str,
        working_directory: Optional[str] = None,
    ) -> int:
        """Create and initialize a new project.

        Uses the InitializerAgent to:
        1. Analyze requirements
        2. Generate comprehensive feature list
        3. Set up project directory
        4. Initialize git repository
        5. Create initial progress log

        Args:
            name: Project name
            description: Detailed requirements/description
            working_directory: Optional directory for project files

        Returns:
            Project ID

        Raises:
            ValueError: If initialization fails
        """
        logger.info(f"Creating project: {name}")

        result = self._initializer.initialize_project(
            name=name,
            description=description,
            working_directory=working_directory,
        )

        if result.get("error"):
            raise ValueError(f"Project initialization failed: {result['error']}")

        project_id = result["project_id"]
        feature_count = result["feature_count"]

        logger.info(
            f"Project {project_id} created with {feature_count} features"
        )

        if self._on_progress:
            self._on_progress(
                {
                    "event": "project_created",
                    "project_id": project_id,
                    "feature_count": feature_count,
                }
            )

        return project_id

    def run_session(
        self,
        project_id: int,
        feature_id: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Run a single working session on a project.

        Each session:
        1. Reviews current state
        2. Selects ONE feature to work on
        3. Implements/fixes the feature
        4. Tests and verifies
        5. Commits and logs progress

        Args:
            project_id: Project to work on
            feature_id: Optional specific feature to work on

        Returns:
            Session result dict with:
            - session_id: ID of the session
            - feature_id: Feature that was worked on
            - verification_result: "passed" or "failed"
            - files_changed: List of modified files
            - error: Any error that occurred
        """
        logger.info(f"Running session for project {project_id}")

        # Check project status
        project = self._project_manager.get_project(project_id)
        if not project:
            return {"error": f"Project {project_id} not found"}

        if project.status == ProjectStatus.COMPLETED:
            return {
                "message": "Project already completed",
                "status": "completed",
            }

        if project.status == ProjectStatus.ABANDONED:
            return {
                "error": "Project has been abandoned",
                "status": "abandoned",
            }

        # Run the session
        result = self._session_agent.run_session(project_id)

        if self._on_progress:
            self._on_progress(
                {
                    "event": "session_complete",
                    "project_id": project_id,
                    **result,
                }
            )

        return result

    def run_until_complete(
        self,
        project_id: int,
        max_sessions: int = 100,
        pause_between_sessions: bool = False,
    ) -> Dict[str, Any]:
        """Run sessions until project is complete or limit reached.

        This is the "autonomous mode" - the harness will continue
        running sessions until all features pass or the session
        limit is reached.

        Args:
            project_id: Project to work on
            max_sessions: Maximum sessions to run (safety limit)
            pause_between_sessions: If True, require confirmation
                between sessions

        Returns:
            Final result dict with:
            - status: "completed", "incomplete", or "error"
            - sessions_run: Number of sessions executed
            - features_passing: Number of passing features
            - total_features: Total feature count
            - error: Any final error
        """
        logger.info(
            f"Running until complete: project {project_id}, max {max_sessions}"
        )

        sessions_run = 0
        errors = []

        for i in range(max_sessions):
            # Check if project is complete
            project = self._project_manager.get_project(project_id)
            if not project:
                return {
                    "status": "error",
                    "error": "Project not found",
                    "sessions_run": sessions_run,
                }

            if project.status == ProjectStatus.COMPLETED:
                logger.info(
                    f"Project completed after {sessions_run} sessions"
                )
                return {
                    "status": "completed",
                    "sessions_run": sessions_run,
                    "features_passing": project.passing_features,
                    "total_features": project.total_features,
                }

            # Run session
            sessions_run += 1
            logger.info(f"Starting session {sessions_run}/{max_sessions}")

            if self._on_progress:
                self._on_progress(
                    {
                        "event": "session_starting",
                        "session_number": sessions_run,
                        "max_sessions": max_sessions,
                        "project_id": project_id,
                    }
                )

            result = self.run_session(project_id)

            if result.get("error"):
                errors.append(result["error"])
                logger.warning(f"Session error: {result['error']}")

            # Check for repeated failures
            if len(errors) >= 5:
                consecutive_errors = errors[-5:]
                if len(set(consecutive_errors)) == 1:
                    logger.error("Same error 5 times in a row, stopping")
                    return {
                        "status": "error",
                        "error": f"Repeated error: {errors[-1]}",
                        "sessions_run": sessions_run,
                        "features_passing": project.passing_features,
                        "total_features": project.total_features,
                    }

            if pause_between_sessions:
                # In a real implementation, this would wait for user input
                logger.info("Session complete. Waiting for next session...")

        # Hit max sessions
        project = self._project_manager.get_project(project_id)
        return {
            "status": "incomplete",
            "message": f"Reached max sessions ({max_sessions})",
            "sessions_run": sessions_run,
            "features_passing": project.passing_features if project else 0,
            "total_features": project.total_features if project else 0,
        }

    def resume_project(self, project_id: int) -> Dict[str, Any]:
        """Resume work on a paused or interrupted project.

        Recovers state from the last session and continues work.

        Args:
            project_id: Project to resume

        Returns:
            Dict with recovery info and first session result
        """
        logger.info(f"Resuming project {project_id}")

        project = self._project_manager.get_project(project_id)
        if not project:
            return {"error": "Project not found"}

        # Get last session for context
        last_session = self._project_manager.get_last_session(project_id)
        recovery_info = None
        if last_session:
            recovery_info = last_session.get_context_for_next_session()
            logger.info(f"Recovered context from session {last_session.id}")

        # Update status if paused
        if project.status == ProjectStatus.PAUSED:
            self._project_manager.update_project_status(
                project_id, ProjectStatus.ACTIVE
            )

        # Run a session
        result = self.run_session(project_id)
        result["recovery_info"] = recovery_info

        return result

    def get_project_status(self, project_id: int) -> Dict[str, Any]:
        """Get current project status and progress.

        Args:
            project_id: Project ID

        Returns:
            Dict with status, progress, and feature breakdown
        """
        project = self._project_manager.get_project(project_id)
        if not project:
            return {"error": "Project not found"}

        features = self._project_manager.get_project_features(project_id)

        # Count by status
        status_counts = {}
        for feature in features:
            status = feature.status.value if feature.status else "not_started"
            status_counts[status] = status_counts.get(status, 0) + 1

        # Recent progress
        progress_log = self._project_manager.get_progress_log(project_id, 5)

        return {
            "project_id": project_id,
            "name": project.name,
            "status": project.status.value if project.status else "unknown",
            "total_features": project.total_features,
            "passing_features": project.passing_features,
            "progress_percent": (
                (project.passing_features / project.total_features * 100)
                if project.total_features
                else 0
            ),
            "feature_breakdown": status_counts,
            "recent_progress": [
                {
                    "timestamp": str(entry.timestamp),
                    "action": entry.action,
                    "outcome": entry.outcome,
                }
                for entry in progress_log
            ],
        }

    def pause_project(self, project_id: int) -> bool:
        """Pause a project.

        Args:
            project_id: Project to pause

        Returns:
            True if paused successfully
        """
        self._project_manager.update_project_status(
            project_id, ProjectStatus.PAUSED
        )
        logger.info(f"Paused project {project_id}")
        return True

    def abandon_project(self, project_id: int, reason: str = "") -> bool:
        """Abandon a project.

        Args:
            project_id: Project to abandon
            reason: Optional reason for abandonment

        Returns:
            True if abandoned successfully
        """
        self._project_manager.update_project_status(
            project_id, ProjectStatus.ABANDONED
        )
        self._project_manager.log_progress(
            project_id=project_id,
            action="Project abandoned",
            outcome=reason or "User requested abandonment",
        )
        logger.info(f"Abandoned project {project_id}: {reason}")
        return True

    def revert_to_checkpoint(
        self, project_id: int, commit_hash: str
    ) -> bool:
        """Revert project to a specific git commit.

        Useful for recovering from bad changes or failed experiments.

        Args:
            project_id: Project ID
            commit_hash: Git commit to revert to

        Returns:
            True if successful
        """
        success = self._project_manager.git_revert_to_commit(
            project_id, commit_hash
        )
        if success:
            self._project_manager.log_progress(
                project_id=project_id,
                action=f"Reverted to commit {commit_hash[:7]}",
                outcome="Recovered previous state",
            )
        return success

    def get_decision_history(
        self, project_id: int, limit: int = 20
    ) -> List[Dict[str, Any]]:
        """Get decision history for learning insights.

        Args:
            project_id: Project ID
            limit: Maximum decisions to return

        Returns:
            List of decision dicts with outcomes
        """
        decisions = self._project_manager.get_relevant_decisions(
            project_id, limit=limit
        )
        return [
            {
                "id": d.id,
                "timestamp": str(d.timestamp),
                "context": d.decision_context,
                "decision": d.decision_made,
                "reasoning": d.reasoning,
                "outcome": d.outcome.value if d.outcome else None,
                "score": d.outcome_score,
                "lesson": d.lesson_learned,
            }
            for d in decisions
        ]

    def add_decision_feedback(
        self,
        decision_id: int,
        outcome: str,
        score: float,
        lesson: Optional[str] = None,
    ) -> None:
        """Add feedback for a past decision.

        This allows humans to provide feedback on agent decisions,
        helping improve future performance.

        Args:
            decision_id: Decision ID
            outcome: "success", "partial", "failure", or "reverted"
            score: -1.0 to 1.0 success score
            lesson: What can be learned
        """
        outcome_enum = DecisionOutcome(outcome)
        self._project_manager.update_decision_outcome(
            decision_id=decision_id,
            outcome=outcome_enum,
            score=score,
            lesson=lesson,
        )
        logger.info(f"Added feedback for decision {decision_id}: {outcome}")

    def export_project_report(self, project_id: int) -> str:
        """Export a comprehensive project report.

        Args:
            project_id: Project ID

        Returns:
            Markdown-formatted report
        """
        project = self._project_manager.get_project(project_id)
        if not project:
            return "# Error: Project not found"

        features = self._project_manager.get_project_features(project_id)
        progress_log = self._project_manager.get_progress_log(project_id, 50)
        git_log = self._project_manager.get_git_log(project_id, 20)
        decisions = self._project_manager.get_relevant_decisions(
            project_id, limit=20
        )

        lines = [
            f"# Project Report: {project.name}",
            "",
            f"**Status:** {project.status.value if project.status else 'unknown'}",
            f"**Created:** {project.created_at}",
            f"**Last Updated:** {project.updated_at}",
            "",
            "## Description",
            project.description or "No description",
            "",
            f"## Progress: {project.get_progress_summary()}",
            "",
            "### Feature Status",
        ]

        # Features by status
        for status in FeatureStatus:
            status_features = [
                f for f in features if f.status == status
            ]
            if status_features:
                lines.append(f"\n#### {status.value.upper()}")
                for f in status_features:
                    lines.append(f"- {f.name}")

        lines.extend(
            [
                "",
                "## Progress Log",
                "",
            ]
        )
        for entry in progress_log:
            lines.append(entry.to_log_string())

        lines.extend(
            [
                "",
                "## Git History",
                "",
            ]
        )
        for commit in git_log:
            lines.append(
                f"- [{commit['hash'][:7]}] {commit['message']}"
            )

        lines.extend(
            [
                "",
                "## Decision History",
                "",
            ]
        )
        for d in decisions:
            outcome = d.outcome.value if d.outcome else "pending"
            lines.append(
                f"- **{d.decision_made}** ({outcome})\n"
                f"  Context: {d.decision_context[:100]}...\n"
                f"  Lesson: {d.lesson_learned or 'None'}"
            )

        return "\n".join(lines)



"""Canvas filter operations mixin.

This mixin provides image filter application, preview, and cancellation
functionality for the canvas scene.
"""

from typing import Any, Optional

from PIL import Image


class CanvasFilterMixin:
    """Mixin for canvas image filter operations.

    Handles applying filters to the active image layer, previewing filters
    before committing, and cancelling filter previews.
    """

    def on_apply_filter_signal(self, message: Any) -> None:
        """Handle filter application signal.

        Args:
            message: Filter object or dict with 'filter_object' key containing
                the filter instance to apply.
        """
        if isinstance(message, dict) and "filter_object" in message:
            filter_object = message["filter_object"]
        else:
            filter_object = message
        self._apply_filter(filter_object)

    def on_cancel_filter_signal(self) -> None:
        """Handle filter cancellation signal.

        Restores the image to its pre-filter state if a preview was active.
        """
        image = self._cancel_filter()
        if image:
            self._load_image_from_object(image=image)

    def on_preview_filter_signal(self, message: dict) -> None:
        """Handle filter preview signal.

        Args:
            message: Dict with 'filter_object' key containing the filter to
                preview.
        """
        filter_object = message["filter_object"]
        filtered_image = self._preview_filter(
            self.current_active_image, filter_object
        )
        self._load_image_from_object(image=filtered_image)

    def _apply_filter(self, filter_object: Any) -> None:
        """Apply a filter to the current active image.

        Only works when drawing_pad_settings is active. Preserves undo state
        and commits changes to history.

        Args:
            filter_object: Filter instance with a filter() method that accepts
                a PIL Image and returns a filtered PIL Image.
        """
        if hasattr(self, "logger") and self.logger:
            self.logger.info(
                f"_apply_filter called with filter: {filter_object}"
            )
            self.logger.info(f"Settings key: {self.settings_key}")

        if self.settings_key != "drawing_pad_settings":
            if hasattr(self, "logger") and self.logger:
                self.logger.warning(
                    f"Filter not applied - wrong settings_key: {self.settings_key}"
                )
            return
        try:
            layer_id = self._add_image_to_undo()
            current = self.current_active_image
            if hasattr(self, "logger") and self.logger:
                self.logger.info(
                    f"Current active image: {current.size if current else None}"
                )
            if current is None:
                if hasattr(self, "logger") and self.logger:
                    self.logger.warning("No current active image to filter")
                return

            try:
                if hasattr(self, "logger") and self.logger:
                    self.logger.info(f"Applying filter to image...")
                filtered_image = filter_object.filter(current)
                if hasattr(self, "logger") and self.logger:
                    self.logger.info(
                        f"Filter applied successfully, result: {filtered_image.size if filtered_image else None}"
                    )
            except Exception:
                if hasattr(self, "logger") and self.logger:
                    self.logger.exception("Filter application failed")
                return

            self.previewing_filter = False
            self.image_backup = None
            self._load_image_from_object(image=filtered_image)
            self._flush_pending_image()
            self._commit_layer_history_transaction(layer_id, "image")
            if hasattr(self, "logger") and self.logger:
                self.logger.info("Filter applied and committed to history")
        except Exception:
            if hasattr(self, "logger") and self.logger:
                self.logger.exception("Unexpected error applying filter")

    def _cancel_filter(self) -> Optional[Image.Image]:
        """Cancel filter preview and restore backup image.

        Returns:
            Backup image if one exists, None otherwise.
        """
        image = None
        if self.image_backup:
            image = self.image_backup.copy()
            self.image_backup = None
        self.previewing_filter = False
        return image

    def _preview_filter(
        self, image: Optional[Image.Image], filter_object: Any
    ) -> Optional[Image.Image]:
        """Preview a filter on the current image without applying it.

        Creates a backup of the original image on first preview, allowing
        the user to cancel and restore. Subsequent previews reuse the backup.

        Args:
            image: PIL Image to preview filter on.
            filter_object: Filter instance with a filter() method.

        Returns:
            Filtered PIL Image, or None if preview is not applicable.
        """
        if self.settings_key != "drawing_pad_settings":
            return None
        if not image:
            return None

        if not self.previewing_filter:
            self.image_backup = image.copy()
            self.previewing_filter = True
        else:
            image = self.image_backup.copy()

        filtered_image = filter_object.filter(image)
        return filtered_image



"""
Model utility functions for AI Runner.

This module provides utilities for determining the correct model storage path for Stable Diffusion and related models.

Functions:
    get_stable_diffusion_model_storage_path(filename: str) -> str
        Returns the absolute path where a model file should be stored in the AI Runner folder.

"""

import os


def get_stable_diffusion_model_storage_path(
    version: str, pipeline_action: str, filename: str
) -> str:
    """Return the absolute path for storing a Stable Diffusion model file in the AI Runner folder.

    Args:
        version (str): Model version (e.g., 'SDXL 1.0').
        pipeline_action (str): Pipeline action (e.g., 'txt2img').
        filename (str): The name of the model file (e.g., 'model.safetensors').

    Returns:
        str: The absolute path where the model should be stored.
    """
    base_dir = os.path.expanduser(
        os.path.join(
            "~",
            ".local",
            "share",
            "airunner",
            "art",
            "models",
            version,
            pipeline_action,
        )
    )
    os.makedirs(base_dir, exist_ok=True)
    return os.path.join(base_dir, filename)



import logging
import socket

from airunner.settings import LOCAL_SERVER_HOST


class NoInternetSocket(socket.socket):
    """
    A custom socket class that prevents any form of internet connections.
    
    When allowed_port is set to a valid port (1-65535), allows connections
    only to localhost on that specific port.
    
    When allowed_port is set to -1 or invalid, blocks ALL connections.
    """
    
    # Class-level attribute to track if port was set
    allowed_port = None
    _block_all = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Check if we should block all connections
        if self.__class__.allowed_port == -1:
            self.__class__._block_all = True
        elif not hasattr(self.__class__, "allowed_port") or self.__class__.allowed_port is None:
            # Port not set - block all connections for safety
            self.__class__._block_all = True
        elif not (1 <= self.__class__.allowed_port <= 65535):
            # Invalid port - block all connections
            self.__class__._block_all = True
        else:
            self.__class__._block_all = False
        self.__init_logger()

    def __init_logger(self):
        self.logger = logging.getLogger(__name__)
        # Use StreamHandler instead of FileHandler to avoid file system issues
        if not self.logger.handlers:
            handler = logging.StreamHandler()
            formatter = logging.Formatter(
                "%(asctime)s - %(levelname)s - %(message)s"
            )
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)

    def connect(self, address):
        """
        Overrides the socket's `connect` method to allow connections only to localhost on a specific port.

        Args:
            address (tuple): A tuple of (host, port)

        Raises:
            ConnectionError: If the connection attempt is to an address other than localhost on the allowed port.
        """
        # Block ALL connections if configured to do so
        if self.__class__._block_all:
            self.logger.warning(
                f"Blocked connection attempt to {address} (all network access disabled)."
            )
            raise ConnectionError(
                f"All network connections are blocked. Connection to {address} is not allowed."
            )
        
        host, port = address
        # Allow localhost connections on the allowed port only
        if host in (LOCAL_SERVER_HOST, "127.0.0.1", "localhost") and port == self.__class__.allowed_port:
            super().connect(address)
            self.logger.info(f"Allowed connection to {host} on port {port}.")
        else:
            self.logger.warning(
                f"Blocked connection attempt to {host} on port {port}."
            )
            raise ConnectionError(
                f"Connection to {host} on port {port} is not allowed. Only localhost:{self.__class__.allowed_port} is permitted."
            )

    @classmethod
    def set_allowed_port(cls, port):
        """
        Set the allowed port for connections.
        
        Args:
            port: Port number (1-65535) to allow, or -1 to block all connections.
        """
        cls.allowed_port = port
        cls._block_all = (port == -1 or port is None or not (1 <= port <= 65535))



"""Initializer Agent for long-running projects.

The Initializer Agent is responsible for setting up a new project:
1. Expanding user requirements into a comprehensive feature list
2. Setting up project directory structure
3. Creating initial configuration files
4. Initializing git repository
5. Writing first progress log entry

This follows Anthropic's pattern but adds:
- More intelligent feature decomposition
- Category-based feature organization
- Dependency detection between features
- Initial decision memory seeding
"""

from typing import Any, Optional, List, Dict, Annotated
from typing_extensions import TypedDict

from langchain_core.messages import BaseMessage, HumanMessage, AIMessage
from langchain_core.prompts import ChatPromptTemplate
from langgraph.graph import START, END, StateGraph, add_messages

from airunner.components.llm.long_running.data.project_state import (
    ProjectStatus,
    FeatureCategory,
)
from airunner.components.llm.long_running.project_manager import ProjectManager
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


# System prompt for the initializer agent
INITIALIZER_SYSTEM_PROMPT = """You are a project initialization specialist. Your task is to analyze a user's project requirements and create a comprehensive, structured feature list.

CRITICAL RULES:
1. Break down the project into ATOMIC features - each should be independently implementable and testable
2. Aim for 20-200 features depending on project complexity
3. Each feature must have clear verification steps
4. Identify dependencies between features
5. Categorize features appropriately
6. Features should be prioritized (1-10, higher = more important)
7. Start with foundational features before UI polish

OUTPUT FORMAT:
You must output a JSON array of features. Each feature:
{
    "name": "Short descriptive name",
    "description": "Detailed description of what this feature does",
    "category": "functional|ui|integration|testing|documentation|performance|security",
    "verification_steps": ["Step 1", "Step 2", "Step 3"],
    "priority": 1-10,
    "depends_on_names": ["Feature Name 1", "Feature Name 2"]  // Names of features this depends on
}

CATEGORIES:
- functional: Core application functionality
- ui: User interface elements
- integration: External service integrations
- testing: Test coverage and validation
- documentation: Documentation and comments
- performance: Optimization and efficiency
- security: Security features and validation

EXAMPLE for a "Todo App":
[
    {
        "name": "Data model for tasks",
        "description": "Create SQLite database schema for storing tasks with id, title, description, completed status, created_at, updated_at fields",
        "category": "functional",
        "verification_steps": ["Database file created", "Schema matches spec", "Can insert/query tasks"],
        "priority": 10,
        "depends_on_names": []
    },
    {
        "name": "Create task API endpoint",
        "description": "POST /tasks endpoint that accepts JSON body with title and description, creates task in database",
        "category": "functional",
        "verification_steps": ["POST creates task", "Returns 201 with task JSON", "Task persisted in DB"],
        "priority": 9,
        "depends_on_names": ["Data model for tasks"]
    }
]

Be thorough but practical. Focus on features that matter for a working application."""


class InitializerState(TypedDict):
    """State schema for Initializer Agent.

    Attributes:
        messages: Conversation messages
        project_name: Name of the project
        project_description: User's requirements
        features_json: Generated feature list as JSON string
        project_id: Created project ID
        error: Any error message
    """

    messages: Annotated[list[BaseMessage], add_messages]
    project_name: str
    project_description: str
    working_directory: Optional[str]
    features_json: Optional[str]
    project_id: Optional[int]
    error: Optional[str]


class InitializerAgent:
    """Agent that initializes long-running projects.

    Takes user requirements and creates:
    1. Project database entry
    2. Comprehensive feature list
    3. Git repository
    4. Initial progress log

    Example:
        ```python
        agent = InitializerAgent(chat_model)
        result = agent.initialize_project(
            name="My Web App",
            description="Build a chat application with user auth and real-time messaging",
            working_directory="/home/user/projects/my-web-app"
        )
        print(f"Created project {result['project_id']} with {result['feature_count']} features")
        ```
    """

    def __init__(
        self,
        chat_model: Any,
        project_manager: Optional[ProjectManager] = None,
    ):
        """Initialize the Initializer Agent.

        Args:
            chat_model: LangChain chat model
            project_manager: ProjectManager instance (creates new if None)
        """
        self._chat_model = chat_model
        self._project_manager = project_manager or ProjectManager()
        self._graph = self._build_graph()

        logger.info("InitializerAgent initialized")

    def _build_graph(self) -> Any:
        """Build the LangGraph workflow for initialization."""
        workflow = StateGraph(InitializerState)

        # Add nodes
        workflow.add_node("analyze_requirements", self._analyze_requirements)
        workflow.add_node("generate_features", self._generate_features)
        workflow.add_node("create_project", self._create_project)
        workflow.add_node("finalize", self._finalize)

        # Add edges
        workflow.add_edge(START, "analyze_requirements")
        workflow.add_edge("analyze_requirements", "generate_features")
        workflow.add_edge("generate_features", "create_project")
        workflow.add_edge("create_project", "finalize")
        workflow.add_edge("finalize", END)

        return workflow.compile()

    def _analyze_requirements(self, state: InitializerState) -> dict:
        """Analyze user requirements and prepare for feature generation.

        Args:
            state: Current state

        Returns:
            Updated state with analysis message
        """
        logger.info(f"Analyzing requirements for project: {state['project_name']}")

        # Create the prompt for analysis
        prompt = ChatPromptTemplate.from_messages(
            [
                ("system", INITIALIZER_SYSTEM_PROMPT),
                (
                    "human",
                    """Analyze these project requirements and prepare a comprehensive feature list.

PROJECT NAME: {project_name}

REQUIREMENTS:
{project_description}

First, identify:
1. Core functionality needed
2. User-facing features
3. Backend/infrastructure needs
4. Testing requirements
5. Potential integrations

Then generate the full feature list in JSON format.""",
                ),
            ]
        )

        # Format the prompt
        formatted = prompt.format_messages(
            project_name=state["project_name"],
            project_description=state["project_description"],
        )

        return {
            "messages": formatted,
        }

    def _generate_features(self, state: InitializerState) -> dict:
        """Generate the feature list using LLM.

        Args:
            state: Current state

        Returns:
            Updated state with features_json
        """
        logger.info("Generating feature list with LLM")

        try:
            # Get LLM response
            response = self._chat_model.invoke(state["messages"])

            # Extract JSON from response
            content = response.content
            features_json = self._extract_json_from_response(content)

            if not features_json:
                return {
                    "error": "Failed to extract feature list from LLM response",
                    "messages": [response],
                }

            return {
                "features_json": features_json,
                "messages": [response],
            }

        except Exception as e:
            logger.error(f"Feature generation failed: {e}")
            return {"error": str(e)}

    def _extract_json_from_response(self, content: str) -> Optional[str]:
        """Extract JSON array from LLM response.

        Args:
            content: LLM response content

        Returns:
            JSON string or None
        """
        import json
        import re

        # Try to find JSON array in response
        # Look for content between [ and ]
        json_match = re.search(r"\[[\s\S]*\]", content)
        if json_match:
            try:
                # Validate it's valid JSON
                json.loads(json_match.group())
                return json_match.group()
            except json.JSONDecodeError:
                pass

        # Try parsing the whole content as JSON
        try:
            json.loads(content)
            return content
        except json.JSONDecodeError:
            pass

        return None

    def _create_project(self, state: InitializerState) -> dict:
        """Create the project and features in database.

        Args:
            state: Current state

        Returns:
            Updated state with project_id
        """
        logger.info(f"Creating project: {state['project_name']}")

        if state.get("error"):
            return {}

        try:
            import json

            # Create the project
            project = self._project_manager.create_project(
                name=state["project_name"],
                description=state["project_description"],
                working_directory=state.get("working_directory"),
                system_prompt=INITIALIZER_SYSTEM_PROMPT,
                init_git=True,
            )

            # Parse and create features
            features_data = json.loads(state["features_json"])

            # First pass: create all features without dependencies
            feature_name_to_id = {}
            for f in features_data:
                feature = self._project_manager.add_feature(
                    project_id=project.id,
                    name=f["name"],
                    description=f.get("description", ""),
                    category=FeatureCategory(f.get("category", "functional")),
                    verification_steps=f.get("verification_steps", []),
                    priority=f.get("priority", 5),
                    depends_on=[],  # Will update in second pass
                )
                feature_name_to_id[f["name"]] = feature.id

            # Second pass: update dependencies
            for f in features_data:
                depends_on_names = f.get("depends_on_names", [])
                if depends_on_names:
                    feature_id = feature_name_to_id.get(f["name"])
                    if feature_id:
                        depends_on_ids = [
                            feature_name_to_id[name]
                            for name in depends_on_names
                            if name in feature_name_to_id
                        ]
                        if depends_on_ids:
                            # Update feature with dependencies
                            feature = self._project_manager.get_feature(feature_id)
                            if feature:
                                feature.depends_on = depends_on_ids

            # Update project status
            self._project_manager.update_project_status(
                project.id, ProjectStatus.ACTIVE
            )

            # Log initial progress
            self._project_manager.log_progress(
                project_id=project.id,
                action="Project initialized",
                outcome=f"Created project with {len(features_data)} features",
                git_commit=True,
            )

            return {
                "project_id": project.id,
            }

        except Exception as e:
            logger.error(f"Project creation failed: {e}")
            return {"error": str(e)}

    def _finalize(self, state: InitializerState) -> dict:
        """Finalize initialization and return results.

        Args:
            state: Current state

        Returns:
            Final state
        """
        if state.get("error"):
            logger.error(f"Initialization failed: {state['error']}")
        else:
            logger.info(
                f"Project {state.get('project_id')} initialized successfully"
            )

        return {}

    def initialize_project(
        self,
        name: str,
        description: str,
        working_directory: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Initialize a new long-running project.

        Args:
            name: Project name
            description: User's requirements/description
            working_directory: Optional directory for project files

        Returns:
            Dict with project_id, feature_count, or error
        """
        logger.info(f"Starting project initialization: {name}")

        initial_state: InitializerState = {
            "messages": [],
            "project_name": name,
            "project_description": description,
            "working_directory": working_directory,
            "features_json": None,
            "project_id": None,
            "error": None,
        }

        result = self._graph.invoke(initial_state)

        if result.get("error"):
            return {"error": result["error"]}

        # Get feature count
        project = self._project_manager.get_project(result["project_id"])
        feature_count = project.total_features if project else 0

        return {
            "project_id": result["project_id"],
            "feature_count": feature_count,
            "project_name": name,
            "status": "initialized",
        }

    def get_feature_list_prompt(self, description: str) -> str:
        """Get a prompt for generating feature list (for testing/debugging).

        Args:
            description: Project description

        Returns:
            Formatted prompt string
        """
        return f"""{INITIALIZER_SYSTEM_PROMPT}

PROJECT REQUIREMENTS:
{description}

Generate the comprehensive feature list in JSON format."""



from airunner.components.nodegraph.gui.widgets.nodes.math.base_math_node import (
    BaseMathNode,
)
from airunner.settings import AIRUNNER_MAX_SEED


class MaxRND(BaseMathNode):
    NODE_NAME = "Max RND Number"
    has_exec_in_port = False
    has_exec_out_port = False
    _input_ports = []
    _output_ports = [
        dict(name="max_seed", display_name="Max Seed"),
    ]
    _properties = []

    def execute(self, input_data):
        return {"max_seed": AIRUNNER_MAX_SEED}



from sqlalchemy import Column, Integer, String

from airunner.components.data.models.base import BaseModel
from airunner.settings import (
    AIRUNNER_DEFAULT_BRUSH_PRIMARY_COLOR,
    AIRUNNER_DEFAULT_BRUSH_SECONDARY_COLOR,
)


class BrushSettings(BaseModel):
    __tablename__ = "brush_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    size = Column(Integer, default=75)
    primary_color = Column(
        String, default=AIRUNNER_DEFAULT_BRUSH_PRIMARY_COLOR
    )
    secondary_color = Column(
        String, default=AIRUNNER_DEFAULT_BRUSH_SECONDARY_COLOR
    )
    strength_slider = Column(Integer, default=950)

    # Relationship to CanvasLayer temporarily commented out
    # TODO: Re-enable after fixing SQLAlchemy relationship mapping
    # layer = relationship("CanvasLayer", back_populates="brush_settings")
    strength = Column(Integer, default=950)
    conditioning_scale = Column(Integer, default=550)
    guidance_scale = Column(Integer, default=75)



from airunner.components.llm.managers.agent import LocalAgent


class PygameAgent(LocalAgent):
    """
    This is an example of a Pygame agent that can be used with the Pygame window.
    It inherits from the LocalAgent class and implements the system_prompt property.
    The system_prompt is used to provide the agent with a specific instruction set
    for interacting with the Pygame window.
    """

    @property
    def system_prompt(self) -> str:
        return (
            """You are a game AI that can interact with the Pygame window."""
        )



"""Base class for all LangGraph visual nodes.

This module provides the base node class that all LangGraph nodes inherit from,
providing common functionality for code generation and validation.
"""

from typing import Dict, Any, List
from airunner.components.nodegraph.gui.widgets.nodes.core.base_workflow_node import (
    BaseWorkflowNode,
)


class BaseLangGraphNode(BaseWorkflowNode):
    """Base class for all LangGraph nodes.

    This class extends BaseWorkflowNode to add LangGraph-specific
    functionality like code generation and state management.

    Attributes:
        __identifier__: Namespace identifier for the node
        state_key: Key in state dict this node modifies
        is_conditional: Does this create a branch?
        is_loop: Does this create a cycle?
        node_color: Visual color for the node (LangGraph nodes are blue-ish)
    """

    __identifier__ = "airunner.langgraph"

    # LangGraph-specific attributes
    state_key: str = ""
    is_conditional: bool = False
    is_loop: bool = False

    # Visual styling for LangGraph nodes
    node_color = (100, 150, 255)  # Blue color to distinguish from other nodes

    def __init__(self):
        """Initialize LangGraph node."""
        super().__init__()

    def to_langgraph_code(self) -> str:
        """Generate LangGraph Python code for this node.

        This method must be implemented by subclasses to generate
        the appropriate Python code for their functionality.

        Returns:
            Python code string

        Raises:
            NotImplementedError: If subclass doesn't implement
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement to_langgraph_code"
        )

    def get_node_config(self) -> Dict[str, Any]:
        """Get node configuration for code generation.

        Returns:
            Dict with node configuration
        """
        return {
            "name": self.name(),
            "type": self.get_node_type(),
            "description": self.get_description(),
            "properties": self.get_all_properties(),
            "state_key": self.state_key,
            "is_conditional": self.is_conditional,
            "is_loop": self.is_loop,
        }

    def get_node_type(self) -> str:
        """Get the node type identifier.

        Returns:
            Node type string (override in subclasses)
        """
        return "custom"

    def get_description(self) -> str:
        """Get node description for generated code.

        Returns:
            Description string
        """
        return f"{self.NODE_NAME} node"

    def get_all_properties(self) -> Dict[str, Any]:
        """Get all node properties as a dict.

        Returns:
            Dict of property_name: value
        """
        properties = {}
        for prop_name in self.properties.keys():
            try:
                properties[prop_name] = self.get_property(prop_name)
            except Exception:
                continue
        return properties

    def validate_connections(self) -> tuple[bool, str]:
        """Validate node connections are valid for LangGraph.

        Returns:
            Tuple of (is_valid, error_message)
        """
        # Check execution ports
        if self.has_exec_in_port:
            exec_in_ports = [
                p
                for p in self.input_ports()
                if p.name() == self.EXEC_IN_PORT_NAME
            ]
            if exec_in_ports:
                exec_in = exec_in_ports[0]
                if not exec_in.connected_ports():
                    # Entry node doesn't need input connection
                    if not self.is_entry_node():
                        return False, "Execution input must be connected"

        # Conditional nodes must have multiple outputs
        if self.is_conditional:
            exec_out_ports = [
                p
                for p in self.output_ports()
                if p.name() == self.EXEC_OUT_PORT_NAME
            ]
            if len(exec_out_ports) < 2:
                return (
                    False,
                    "Conditional nodes must have at least 2 output paths",
                )

        return True, ""

    def is_entry_node(self) -> bool:
        """Check if this is an entry node (no exec input needed).

        Returns:
            True if entry node, False otherwise
        """
        return False

    def get_state_requirements(self) -> Dict[str, type]:
        """Get required state fields for this node.

        Returns:
            Dict of field_name: type
        """
        return {}

    def get_input_state_keys(self) -> List[str]:
        """Get state keys this node reads from.

        Returns:
            List of state field names
        """
        return []

    def get_output_state_keys(self) -> List[str]:
        """Get state keys this node writes to.

        Returns:
            List of state field names
        """
        if self.state_key:
            return [self.state_key]
        return []

    def execute(self, input_data: Dict[str, Any]) -> Dict[str, Any]:
        """Execute node (for runtime workflow execution).

        Args:
            input_data: Input data from previous nodes

        Returns:
            Output data for next nodes
        """
        # LangGraph nodes are executed via generated code,
        # not through this method
        return input_data

    def get_next_node(self, state: Dict[str, Any]) -> str:
        """Get next node name based on state (for conditional nodes).

        Args:
            state: Current workflow state

        Returns:
            Name of next node to execute
        """
        # Default: follow first output connection
        exec_out_ports = [
            p
            for p in self.output_ports()
            if p.name() == self.EXEC_OUT_PORT_NAME
        ]
        if exec_out_ports and exec_out_ports[0].connected_ports():
            next_port = exec_out_ports[0].connected_ports()[0]
            return next_port.node().name()

        return "END"



"""
ContextManager for managing contextual information for the LLM.
Tracks browser tabs, code files, and other sources by unique key.

Google Python Style Guide applies.
"""

from typing import Dict, Any, Optional


class SingletonMeta(type):
    """Metaclass for implementing singleton pattern."""

    _instances: Dict[type, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class ContextManager(metaclass=SingletonMeta):
    """Manages contextual information for the LLM, including browser tabs, code files, and other sources.

    Context is stored as a dictionary keyed by a unique identifier (e.g., URL, file path).
    """

    def __init__(self) -> None:
        self._contexts: Dict[str, Dict[str, Any]] = {}

    def set_context(self, key: str, context: Dict[str, Any]) -> None:
        """Add or update context for a given key.

        Args:
            key (str): Unique identifier for the context (e.g., URL, file path).
            context (dict): Context data.
        """
        self._contexts[key] = context

    def remove_context(self, key: str) -> None:
        """Remove context for a given key.

        Args:
            key (str): Unique identifier for the context.
        """
        self._contexts.pop(key, None)

    def get_context(self, key: str) -> Optional[Dict[str, Any]]:
        """Retrieve context for a given key.

        Args:
            key (str): Unique identifier for the context.

        Returns:
            dict or None: Context data if present.
        """
        return self._contexts.get(key)

    def all_contexts(self) -> Dict[str, Dict[str, Any]]:
        """Get all current contexts.

        Returns:
            dict: All context data.
        """
        return self._contexts

    def clear(self) -> None:
        """Remove all contexts."""
        self._contexts.clear()



"""
Unit tests for the tool search engine.

Tests BM25 search functionality, fallback search, and tool indexing.
"""

import pytest
from unittest.mock import patch, MagicMock

from airunner.components.llm.core.tool_registry import (
    ToolRegistry,
    ToolInfo,
    ToolCategory,
    tool,
)
from airunner.components.llm.core.tool_search import (
    ToolSearchEngine,
    get_tool_search_engine,
)


@pytest.fixture(autouse=True)
def clear_registry():
    """Clear the registry before and after each test."""
    ToolRegistry.clear()
    yield
    ToolRegistry.clear()


@pytest.fixture
def sample_tools():
    """Register sample tools for testing."""
    @tool(
        name="test_search_web",
        category=ToolCategory.SEARCH,
        description="Search the internet for information",
        defer_loading=True,
        keywords=["internet", "google", "query"],
    )
    def search_web(query: str) -> str:
        return f"Results for {query}"

    @tool(
        name="test_generate_image",
        category=ToolCategory.IMAGE,
        description="Generate an image from text prompt",
        defer_loading=True,
        keywords=["picture", "art", "create", "draw"],
    )
    def generate_image(prompt: str) -> str:
        return f"Image: {prompt}"

    @tool(
        name="test_read_file",
        category=ToolCategory.FILE,
        description="Read contents of a file from disk",
        defer_loading=True,
        keywords=["open", "load", "text", "document"],
    )
    def read_file(path: str) -> str:
        return f"Contents of {path}"

    @tool(
        name="test_immediate_tool",
        category=ToolCategory.SYSTEM,
        description="An always-available tool",
        defer_loading=False,
    )
    def immediate_tool() -> str:
        return "Always here"

    return {
        "search_web": search_web,
        "generate_image": generate_image,
        "read_file": read_file,
        "immediate_tool": immediate_tool,
    }


class TestToolSearchEngine:
    """Tests for ToolSearchEngine class."""

    def test_init_indexes_deferred_tools(self, sample_tools):
        """Engine should index only deferred tools by default."""
        engine = ToolSearchEngine(include_immediate=False)
        
        # Should have indexed 3 deferred tools
        assert len(engine._tools) == 3
        
        # Should not include immediate tool
        tool_names = [t.name for t in engine._tools]
        assert "test_immediate_tool" not in tool_names
        assert "test_search_web" in tool_names

    def test_init_with_immediate(self, sample_tools):
        """Engine can optionally include immediate tools."""
        engine = ToolSearchEngine(include_immediate=True)
        
        # Should have all 4 tools
        assert len(engine._tools) == 4
        
        tool_names = [t.name for t in engine._tools]
        assert "test_immediate_tool" in tool_names

    def test_search_finds_relevant_tools(self, sample_tools):
        """Search should return relevant tools based on query."""
        engine = ToolSearchEngine(include_immediate=False)
        
        results = engine.search("search the internet", limit=5)
        
        assert len(results) > 0
        # Search web tool should be first or near first
        result_names = [t.name for t in results]
        assert "test_search_web" in result_names

    def test_search_respects_limit(self, sample_tools):
        """Search should respect the limit parameter."""
        engine = ToolSearchEngine(include_immediate=False)
        
        results = engine.search("tool", limit=1)
        
        assert len(results) <= 1

    def test_search_with_keywords(self, sample_tools):
        """Search should match on keywords."""
        engine = ToolSearchEngine(include_immediate=False)
        
        # Search for "picture" which is a keyword for generate_image
        results = engine.search("picture", limit=5)
        
        result_names = [t.name for t in results]
        assert "test_generate_image" in result_names

    def test_search_empty_query(self, sample_tools):
        """Search with empty query should return empty results."""
        engine = ToolSearchEngine(include_immediate=False)
        
        results = engine.search("", limit=5)
        
        assert results == []

    def test_search_no_matches(self, sample_tools):
        """Search with unrelated query should return empty results."""
        engine = ToolSearchEngine(include_immediate=False)
        
        # Search for something completely unrelated
        results = engine.search("xyzzyx12345", limit=5)
        
        # May return empty or low-scoring results
        assert len(results) >= 0

    def test_tokenize(self, sample_tools):
        """Tokenizer should lowercase and split on non-alphanumeric."""
        engine = ToolSearchEngine()
        
        tokens = engine._tokenize("Search the Web! 123")
        
        assert "search" in tokens
        assert "the" in tokens
        assert "web" in tokens
        assert "123" in tokens
        # Should be lowercase
        assert "Search" not in tokens

    def test_refresh_updates_index(self, sample_tools):
        """Refresh should update the index with new tools."""
        engine = ToolSearchEngine(include_immediate=False)
        initial_count = len(engine._tools)
        
        # Add a new deferred tool
        @tool(
            name="test_new_tool",
            category=ToolCategory.SYSTEM,
            description="A newly added tool",
            defer_loading=True,
        )
        def new_tool() -> str:
            return "New"
        
        # Refresh the index
        engine.refresh()
        
        assert len(engine._tools) == initial_count + 1

    def test_fallback_search_without_bm25(self, sample_tools):
        """Engine should work without rank_bm25 installed."""
        engine = ToolSearchEngine(include_immediate=False)
        
        # Force fallback mode
        engine._index = None
        
        results = engine._fallback_search(["search", "internet"], limit=5)
        
        assert len(results) > 0
        result_names = [t.name for t in results]
        assert "test_search_web" in result_names


class TestGetToolSearchEngine:
    """Tests for the get_tool_search_engine singleton."""

    def test_returns_same_instance(self, sample_tools):
        """Should return same instance for same parameters."""
        # Reset global
        import airunner.components.llm.core.tool_search as module
        module._search_engine = None
        
        engine1 = get_tool_search_engine(include_immediate=False)
        engine2 = get_tool_search_engine(include_immediate=False)
        
        assert engine1 is engine2

    def test_creates_new_for_different_params(self, sample_tools):
        """Should create new instance for different parameters."""
        import airunner.components.llm.core.tool_search as module
        module._search_engine = None
        
        engine1 = get_tool_search_engine(include_immediate=False)
        engine2 = get_tool_search_engine(include_immediate=True)
        
        assert engine1 is not engine2



from typing import Optional
from alembic import op
import sqlalchemy as sa

from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger


logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


def get_tables():
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    return inspector.get_table_names()


def table_exists(table_name):
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    return table_name in inspector.get_table_names()


def add_table(cls):
    if not table_exists(cls.__tablename__):
        columns = [column.copy() for column in cls.__table__.columns]
        op.create_table(
            cls.__tablename__, *columns, *getattr(cls, "__table_args__", ())
        )
    else:
        logger.warning(
            f"Table '{cls.__tablename__}' already exists, skipping add."
        )
    return


def add_tables(classes):
    for cls in classes:
        create_table_with_defaults(cls)
    return


def drop_table(cls: Optional[object] = None, table_name: Optional[str] = None):
    if cls is not None:
        table_name = cls.__tablename__
    if table_exists(table_name):
        op.drop_table(table_name)
    else:
        logger.warning(f"Table '{table_name}' does not exist, skipping drop.")
    return


def drop_tables(classes):
    for cls in classes:
        drop_table(cls)
    return


def create_table_with_defaults(model):
    if not table_exists(model.__tablename__):
        try:
            columns = []
            for column in model.__table__.columns:
                column_copy = column.copy()
                if column.default is not None:
                    column_copy.server_default = column.default
                columns.append(column_copy)
            op.create_table(
                model.__tablename__,
                *columns,
                *getattr(model, "__table_args__", ()),
            )
            set_default_values(model)
        except Exception as e:
            logger.error(
                f"Failed to create table {model.__tablename__}: {str(e)}"
            )
    else:
        logger.warning(f"{model.__tablename__} already exists, skipping")


def set_default_values(model):
    default_values = {}
    for column in model.__table__.columns:
        if column.default is not None:
            default_values[column.name] = column.default.arg
    op.bulk_insert(model.__table__, [default_values])



from sqlalchemy import JSON, Column, Integer, String

from airunner.components.data.models.base import BaseModel


class Tool(BaseModel):
    __tablename__ = "tools"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    data = Column(String, nullable=False)
    description = Column(String, nullable=True)
    props = Column(JSON, nullable=True, default={})



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'user_agreement.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QGridLayout, QLabel,
    QScrollArea, QSizePolicy, QWidget)

class Ui_user_agreement(object):
    def setupUi(self, user_agreement):
        if not user_agreement.objectName():
            user_agreement.setObjectName(u"user_agreement")
        user_agreement.resize(400, 526)
        self.gridLayout_2 = QGridLayout(user_agreement)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.scrollArea = QScrollArea(user_agreement)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 366, 1565))
        self.gridLayout = QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label_2 = QLabel(self.scrollAreaWidgetContents)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setWordWrap(True)

        self.gridLayout.addWidget(self.label_2, 0, 0, 1, 1)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.gridLayout_2.addWidget(self.scrollArea, 1, 0, 1, 1)

        self.label = QLabel(user_agreement)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setPointSize(11)
        font.setBold(True)
        self.label.setFont(font)

        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)

        self.checkBox = QCheckBox(user_agreement)
        self.checkBox.setObjectName(u"checkBox")

        self.gridLayout_2.addWidget(self.checkBox, 2, 0, 1, 1)


        self.retranslateUi(user_agreement)
        self.checkBox.clicked["bool"].connect(user_agreement.agreement_clicked)

        QMetaObject.connectSlotsByName(user_agreement)
    # setupUi

    def retranslateUi(self, user_agreement):
        user_agreement.setWindowTitle(QCoreApplication.translate("user_agreement", u"Form", None))
        self.label_2.setText(QCoreApplication.translate("user_agreement", u"Loading user agreement...", None))
        self.label.setText(QCoreApplication.translate("user_agreement", u"User Agreement", None))
        self.checkBox.setText(QCoreApplication.translate("user_agreement", u"I have read and agree to the terms of the User Agreement", None))
    # retranslateUi




"""
Custom EPUB reader for EPUBs with numbered page structure
"""

import os
import zipfile
from typing import List, Optional
from bs4 import BeautifulSoup
from llama_index.core import Document
from llama_index.core.readers.base import BaseReader


class CustomEpubReader(BaseReader):
    """Custom EPUB reader that handles numbered page structure"""

    def load_data(
        self, file: Optional[str] = None, **kwargs
    ) -> List[Document]:
        """Load data from EPUB file."""
        if not file or not os.path.exists(file):
            return []

        documents = []

        try:
            with zipfile.ZipFile(file, "r") as zip_file:
                # Get all HTML files in order
                html_files = [
                    f
                    for f in zip_file.namelist()
                    if f.endswith((".html", ".xhtml", ".htm"))
                ]

                # Sort page files numerically if they follow page_X.html pattern
                page_files = [
                    f
                    for f in html_files
                    if "page_" in f and not f.endswith("nav.xhtml")
                ]
                page_files.sort(key=lambda x: self._extract_page_number(x))

                # If no page files, fall back to all HTML files
                if not page_files:
                    page_files = html_files

                all_content = []

                for html_file in page_files:
                    try:
                        with zip_file.open(html_file) as f:
                            content = f.read().decode("utf-8", errors="ignore")
                            soup = BeautifulSoup(content, "html.parser")
                            text = soup.get_text()
                            cleaned_text = " ".join(text.split())

                            if cleaned_text and len(cleaned_text.strip()) > 10:
                                all_content.append(cleaned_text)

                    except Exception as e:
                        print(f"Warning: Could not read {html_file}: {e}")
                        continue

                # Combine all content into a single document
                if all_content:
                    full_text = "\n\n".join(all_content)
                    metadata = {
                        "file_path": file,
                        "file_name": os.path.basename(file),
                        "file_type": "application/epub+zip",
                        "pages_processed": len(all_content),
                    }

                    documents.append(
                        Document(text=full_text, metadata=metadata)
                    )

        except Exception as e:
            print(f"Error reading EPUB {file}: {e}")

        return documents

    def _extract_page_number(self, filename: str) -> int:
        """Extract page number from filename like 'EPUB/page_123.html'"""
        try:
            # Extract number from page_X.html pattern
            parts = filename.split("page_")
            if len(parts) > 1:
                number_part = parts[1].split(".")[0]
                return int(number_part)
        except (ValueError, IndexError):
            pass
        return 0



"""Video encoding mixin for text and image processing."""

import torch
import numpy as np
from PIL import Image
from typing import Tuple, Any


class VideoEncodingMixin:
    """Mixin for encoding text and images for video generation.

    Handles:
    - Text prompt encoding (Llama + CLIP)
    - Image preprocessing and encoding
    - VAE encoding
    - CLIP Vision encoding

    Dependencies (from parent):
        text_encoder: Llama text encoder
        text_encoder_2: CLIP text encoder
        tokenizer: Llama tokenizer
        tokenizer_2: CLIP tokenizer
        vae: Video VAE
        image_encoder: CLIP vision encoder
        feature_extractor: CLIP feature extractor
        high_vram: Boolean for VRAM mode
        logger: Logger instance
        _emit_progress: Progress callback method
    """

    def _encode_text_prompts(
        self,
        prompt: str,
        negative_prompt: str,
        cfg_scale: float,
        gpu: torch.device,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Encode text prompts using Llama and CLIP.

        Args:
            prompt: Main text prompt
            negative_prompt: Negative prompt
            cfg_scale: Classifier-free guidance scale
            gpu: Target device

        Returns:
            Tuple of (llama_vec, llama_attention_mask, llama_vec_n, llama_attention_mask_n,
                     clip_l_pooler, clip_l_pooler_n)
        """
        from airunner.vendor.framepack.diffusers_helper import (
            hunyuan,
            memory as memory_utils,
        )
        from airunner.vendor.framepack.diffusers_helper.utils import (
            crop_or_pad_yield_mask,
        )

        self._emit_progress(10, "Encoding text prompts...")

        if not self.high_vram:
            memory_utils.fake_diffusers_current_device(self.text_encoder, gpu)
            memory_utils.load_model_as_complete(
                self.text_encoder_2, target_device=gpu
            )

        # Encode prompt
        llama_vec, clip_l_pooler = hunyuan.encode_prompt_conds(
            prompt,
            self.text_encoder,
            self.text_encoder_2,
            self.tokenizer,
            self.tokenizer_2,
        )

        # Encode negative prompt
        if cfg_scale == 1:
            llama_vec_n, clip_l_pooler_n = torch.zeros_like(
                llama_vec
            ), torch.zeros_like(clip_l_pooler)
        else:
            llama_vec_n, clip_l_pooler_n = hunyuan.encode_prompt_conds(
                negative_prompt,
                self.text_encoder,
                self.text_encoder_2,
                self.tokenizer,
                self.tokenizer_2,
            )

        # Process text embeddings
        llama_vec, llama_attention_mask = crop_or_pad_yield_mask(
            llama_vec, length=512
        )
        llama_vec_n, llama_attention_mask_n = crop_or_pad_yield_mask(
            llama_vec_n, length=512
        )

        return (
            llama_vec,
            llama_attention_mask,
            llama_vec_n,
            llama_attention_mask_n,
            clip_l_pooler,
            clip_l_pooler_n,
        )

    def _process_and_encode_image(
        self, init_image: Any, height: int, width: int, gpu: torch.device
    ) -> Tuple[torch.Tensor, torch.Tensor, np.ndarray]:
        """Process input image and encode with VAE and CLIP Vision.

        Args:
            init_image: Input PIL Image or numpy array
            height: Target height
            width: Target width
            gpu: Target device

        Returns:
            Tuple of (start_latent, image_encoder_hidden_states, input_image_np)
        """
        from airunner.vendor.framepack.diffusers_helper import (
            hunyuan,
            clip_vision,
            memory as memory_utils,
        )
        from airunner.vendor.framepack.diffusers_helper.utils import (
            resize_and_center_crop,
        )

        self._emit_progress(5, "Processing input image...")

        # Convert to numpy if needed
        if isinstance(init_image, Image.Image):
            input_image_np = np.array(init_image)
        else:
            input_image_np = init_image

        # Resize and crop
        input_image_np = resize_and_center_crop(
            input_image_np, target_width=width, target_height=height
        )

        # Convert to tensor
        input_image_pt = torch.from_numpy(input_image_np).float() / 127.5 - 1
        input_image_pt = input_image_pt.permute(2, 0, 1)[None, :, None]

        # VAE encoding
        self._emit_progress(20, "VAE encoding...")

        if not self.high_vram:
            memory_utils.load_model_as_complete(self.vae, target_device=gpu)

        start_latent = hunyuan.vae_encode(input_image_pt.to(gpu), self.vae)

        # CLIP Vision encoding
        self._emit_progress(30, "Vision encoding...")

        if not self.high_vram:
            memory_utils.load_model_as_complete(
                self.image_encoder, target_device=gpu
            )

        image_encoder_output = clip_vision.hf_clip_vision_encode(
            input_image_np, self.feature_extractor, self.image_encoder
        )
        image_encoder_hidden_states = image_encoder_output.last_hidden_state

        return start_latent, image_encoder_hidden_states, input_image_np



"""drop knowledge_facts and knowledge_relationships tables

Revision ID: f480bbc9acdb
Revises: 9d70f20f2fed
Create Date: 2025-11-28 15:42:28.190695

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'f480bbc9acdb'
down_revision: Union[str, None] = '9d70f20f2fed'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Drop the knowledge tables - we now use markdown-based knowledge base instead
    # Check if tables exist before dropping (SQLite compatibility)
    bind = op.get_bind()
    inspector = sa.inspect(bind)
    existing_tables = inspector.get_table_names()
    
    if 'knowledge_relationships' in existing_tables:
        op.drop_table('knowledge_relationships')
    if 'knowledge_facts' in existing_tables:
        op.drop_table('knowledge_facts')
    if 'conversation_summaries' in existing_tables:
        op.drop_table('conversation_summaries')


def downgrade() -> None:
    # Recreate the tables if needed (not recommended)
    op.create_table(
        'knowledge_facts',
        sa.Column('id', sa.Integer(), primary_key=True, autoincrement=True),
        sa.Column('text', sa.Text(), nullable=False),
        sa.Column('category', sa.String(50), nullable=False, default='other'),
        sa.Column('tags', sa.JSON(), nullable=True),
        sa.Column('confidence', sa.Float(), nullable=False, default=0.9),
        sa.Column('source', sa.String(50), nullable=False, default='conversation'),
        sa.Column('source_conversation_id', sa.Integer(), nullable=True),
        sa.Column('verified', sa.Boolean(), nullable=False, default=False),
        sa.Column('enabled', sa.Boolean(), nullable=False, default=True),
        sa.Column('metadata_json', sa.JSON(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('last_accessed', sa.DateTime(), nullable=True),
        sa.Column('access_count', sa.Integer(), nullable=False, default=0),
    )
    op.create_table(
        'knowledge_relationships',
        sa.Column('id', sa.Integer(), primary_key=True, autoincrement=True),
        sa.Column('source_fact_id', sa.Integer(), nullable=False),
        sa.Column('target_fact_id', sa.Integer(), nullable=True),
        sa.Column('relationship_type', sa.String(50), nullable=False),
        sa.Column('entity_name', sa.String(255), nullable=True),
        sa.Column('entity_type', sa.String(50), nullable=True),
        sa.Column('confidence', sa.Float(), nullable=False, default=0.9),
        sa.Column('metadata_json', sa.JSON(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
    )


from sqlalchemy import Column, Integer, Float

from airunner.components.data.models.base import BaseModel


class STTSettings(BaseModel):
    __tablename__ = "stt_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    duration = Column(Integer, default=10)
    fs = Column(Integer, default=16000)
    channels = Column(Integer, default=1)
    volume_input_threshold = Column(Float, default=0.08)
    silence_buffer_seconds = Column(Float, default=1.0)
    chunk_duration = Column(Float, default=0.03)



"""Canvas Persistence Mixin.

Handles image persistence operations including export, import, load from files/URLs,
and asynchronous image data persistence to the database.
"""

from concurrent.futures import ThreadPoolExecutor
import os
import weakref
from PIL import Image
from PySide6.QtWidgets import QFileDialog
from PySide6.QtCore import Slot
from airunner.utils.image import export_image
from airunner.settings import AIRUNNER_VALID_IMAGE_FILES
from airunner.utils.image.dispatch_persist_result import (
    dispatch_persist_result,
)
from airunner.utils.image.persist_image_worker import persist_image_worker

PERSIST_EXECUTOR = ThreadPoolExecutor(max_workers=1)


class CanvasPersistenceMixin:
    """Provides image persistence functionality to canvas scenes.

    Handles:
    - Export/import images via file dialogs
    - Load images from file paths
    - Asynchronous image data persistence to database
    - Pending image flush mechanism
    """

    def on_export_image_signal(self):
        """Export current active image to a file.

        Opens a file dialog for the user to select export location and format.
        Supports all valid AI Runner image formats.
        """
        image = self.current_active_image
        if image:
            parent_window = self.views()[0].window()
            initial_dir = (
                self.last_export_path if self.last_export_path else ""
            )
            file_dialog = QFileDialog(
                parent_window,
                "Save Image",
                initial_dir,
                f"Image Files ({' '.join(AIRUNNER_VALID_IMAGE_FILES)})",
            )
            file_dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
            if file_dialog.exec() == QFileDialog.DialogCode.Accepted:
                file_path = file_dialog.selectedFiles()[0]
                if file_path == "":
                    return
                self.last_export_path = os.path.dirname(file_path)
                if not file_path.endswith(AIRUNNER_VALID_IMAGE_FILES):
                    file_path = f"{file_path}.png"
                export_image(image, file_path)

    def on_import_image_signal(self):
        """Import an image from a file.

        Opens a file dialog for the user to select an image file to import.
        Only works for drawing pad settings context.
        """
        if self.settings_key != "drawing_pad_settings":
            return
        file_path, _ = QFileDialog.getOpenFileName(
            None,
            "Open Image",
            "",
            f"Image Files ({' '.join(AIRUNNER_VALID_IMAGE_FILES)})",
        )
        if file_path == "":
            return
        self.on_load_image_signal(file_path)

    def on_load_image_from_path(self, message):
        """Load an image from a file path provided in a message dict.

        Args:
            message: Dict containing 'image_path' key
        """
        image_path = message["image_path"]
        if image_path is None or image_path == "":
            return
        image = Image.open(image_path)
        self._load_image_from_object(image)

    def on_load_image_signal(self, image_path: str):
        """Load an image from a file path with undo history support.

        Args:
            image_path: Path to the image file to load
        """
        layer_id = self._add_image_to_undo()
        image = self._load_image(image_path)
        if self.application_settings.resize_on_paste:
            image = self._resize_image(image)
        self.current_active_image = image
        self.initialize_image(image)
        self._commit_layer_history_transaction(layer_id, "image")

    def _flush_pending_image(self):
        """Persist pending image data asynchronously to avoid UI stalls.

        Handles deferred image persistence using a background worker thread.
        Skips persistence if user is actively drawing or if no changes detected.
        """
        if self._is_user_interacting or getattr(
            self, "draw_button_down", False
        ):
            self._persist_timer.start(self._persist_delay_ms)
            return

        has_pending_image = self._pending_image_ref is not None
        has_pending_binary = self._pending_image_binary is not None
        if not has_pending_image and not has_pending_binary:
            return

        image_obj = self._pending_image_ref
        binary = self._pending_image_binary

        self._pending_image_ref = None
        self._pending_image_binary = None

        if binary is not None and binary == self._current_active_image_binary:
            return

        if binary is not None:
            self._current_active_image_binary = binary

        image_payload = None
        if has_pending_image and isinstance(image_obj, Image.Image):
            image_payload = image_obj

        if binary is None and image_payload is None:
            return

        try:
            layer_id = self._get_current_selected_layer_id()
        except Exception:
            layer_id = None

        self._persist_generation += 1
        generation = self._persist_generation

        future = PERSIST_EXECUTOR.submit(
            persist_image_worker,
            self.settings_key,
            layer_id,
            "image",
            image_payload,
            binary,
            self._raw_image_storage_enabled,
            generation,
        )
        self._active_persist_future = future
        future.add_done_callback(
            lambda fut, scene_ref=weakref.ref(self): dispatch_persist_result(
                scene_ref, fut
            )
        )

    @Slot(object)
    def _handle_persist_result(self, payload: object):
        """Handle the result from async image persistence worker.

        Args:
            payload: Result dict from persistence worker containing:
                - generation: Persistence generation number
                - error: Error message if persistence failed
                - binary: Persisted binary data
                - table_name: Database table name
                - column_name: Database column name
        """
        if not isinstance(payload, dict):
            return

        generation = payload.get("generation", 0)
        if generation < self._persist_generation:
            return

        self._active_persist_future = None

        error = payload.get("error")
        if error:
            if hasattr(self, "logger") and self.logger:
                self.logger.error(f"Image persistence failed: {error}")
            return

        binary = payload.get("binary")
        if binary is not None:
            self._current_active_image_binary = binary

        table_name = payload.get("table_name")
        column_name = payload.get("column_name")

        if table_name and column_name:
            try:
                self._notify_setting_updated(
                    setting_name=table_name,
                    column_name=column_name,
                    val=binary,
                )
            except Exception as exc:  # pragma: no cover - defensive
                if hasattr(self, "logger") and self.logger:
                    self.logger.error(
                        f"Failed to notify settings update for {table_name}.{column_name}: {exc}"
                    )

    @staticmethod
    def _load_image(image_path: str) -> Image:
        """Load an image from a file path.

        Args:
            image_path: Path to the image file

        Returns:
            Loaded PIL Image object
        """
        image = Image.open(image_path)
        return image



from sqlalchemy import Column, Integer, Boolean, LargeBinary, ForeignKey
from sqlalchemy.orm import relationship

from airunner.components.data.models.base import BaseModel


class OutpaintSettings(BaseModel):
    __tablename__ = "outpaint_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    layer_id = Column(
        Integer,
        ForeignKey("canvas_layer.id", ondelete="CASCADE"),
        nullable=True,
    )
    image = Column(LargeBinary, nullable=True)
    enabled = Column(Boolean, default=False)
    strength = Column(Integer, default=50)
    mask_blur = Column(Integer, default=0)
    use_grid_image_as_input = Column(Boolean, default=False)
    lock_input_image = Column(Boolean, default=False)

    # Relationship to CanvasLayer temporarily commented out
    # TODO: Re-enable after fixing SQLAlchemy relationship mapping
    # layer = relationship("CanvasLayer", back_populates="outpaint_settings")



"""RAG properties and configuration management.

This mixin provides:
- Property accessors for RAG components
- Configuration path management
- Embedding model setup
- Text splitter configuration
"""

import os
from typing import Optional, Dict, Any
import torch

from llama_index.core.node_parser import SentenceSplitter
from llama_index.embeddings.huggingface import HuggingFaceEmbedding

from airunner.settings import AIRUNNER_LOCAL_FILES_ONLY


class RAGPropertiesMixin:
    """Mixin for RAG properties and configuration."""

    @property
    def text_splitter(self) -> SentenceSplitter:
        """Get or create the text splitter for document chunking.

        Returns:
            Configured SentenceSplitter instance
        """
        if self._text_splitter is None:
            # Get RAG settings with safe defaults
            rag_settings = getattr(self, "rag_settings", None)
            chunk_size = 512  # default
            chunk_overlap = 50  # default

            if rag_settings is not None:
                chunk_size = getattr(rag_settings, "chunk_size", 512)
                chunk_overlap = getattr(rag_settings, "chunk_overlap", 50)

            self._text_splitter = SentenceSplitter(
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap,
            )
        return self._text_splitter

    @property
    def doc_indexes_dir(self) -> str:
        """Get the directory path where per-document indexes are stored.

        Returns:
            Absolute path to doc_indexes directory
        """
        base_dir = os.path.join(
            os.path.expanduser(self.path_settings.base_path),
            "rag",
            "doc_indexes",
        )
        os.makedirs(base_dir, exist_ok=True)
        return base_dir

    @property
    def registry_path(self) -> str:
        """Get the path to the index registry JSON file.

        Returns:
            Absolute path to index_registry.json
        """
        return os.path.join(self.doc_indexes_dir, "index_registry.json")

    @property
    def index_registry(self) -> Dict[str, Any]:
        """Get the index registry (lazy load).

        Returns:
            Dictionary mapping doc_id to metadata
        """
        if self._index_registry is None:
            self._index_registry = self._load_registry()
        return self._index_registry

    @property
    def embedding(self) -> HuggingFaceEmbedding:
        """Get or create the embedding model for RAG.

        This is a lazy-loaded property that creates the embedding model
        only when needed and caches it for reuse.

        Returns:
            HuggingFaceEmbedding instance
        """
        if self._embedding is None:
            try:
                # Construct local path to embedding model files
                # Must use local filesystem path, not HuggingFace repo ID
                model_name = os.path.expanduser(
                    os.path.join(
                        self.path_settings.base_path,
                        "text",
                        "models",
                        "llm",
                        "embedding",
                        "intfloat/e5-large",
                    )
                )

                # Check if model files exist and trigger download if needed
                if not self._check_and_download_embedding_model(model_name):
                    self.logger.warning(
                        "Embedding model files missing - download in progress"
                    )
                    return None

                device = "cpu"
                if torch.cuda.is_available():
                    device = "cuda"
                elif (
                    hasattr(torch.backends, "mps")
                    and torch.backends.mps.is_available()
                ):
                    device = "mps"

                self.logger.info(
                    f"Initializing embedding model on {device}: "
                    f"{model_name}"
                )

                # Set quantization flag if using CPU or limited GPU memory
                # This can reduce memory usage significantly
                # quantization_flag = device == "cpu"

                # Use fp16 to reduce VRAM usage (~650MB vs ~1.3GB)
                model_kwargs = {}
                if device == "cuda":
                    model_kwargs["torch_dtype"] = torch.float16

                self._embedding = HuggingFaceEmbedding(
                    model_name=model_name,
                    device=device,
                    trust_remote_code=True,
                    local_files_only=AIRUNNER_LOCAL_FILES_ONLY,
                    model_kwargs=model_kwargs,
                )

                self.logger.info("Embedding model initialized successfully")

            except Exception as e:
                self.logger.error(
                    f"Failed to initialize embedding model: {e}",
                    exc_info=True,
                )
                # Don't raise - return None to allow RAG setup to skip gracefully
                return None

        return self._embedding

    @property
    def target_files(self) -> Optional[list[str]]:
        """Get the list of target files for indexing.

        Returns:
            List of file paths or None if targeting all documents
        """
        return self._target_files

    @target_files.setter
    def target_files(self, value: Optional[list[str]]):
        """Set the list of target files for indexing.

        Args:
            value: List of file paths or None to target all documents
        """
        self._target_files = value
        if value:
            self.logger.info(f"Set {len(value)} target files for indexing")
        else:
            self.logger.info("Cleared target files (will index all documents)")

    @property
    def rag_system_prompt(self) -> str:
        """Generate system prompt for RAG-enhanced chat.

        Returns:
            System prompt string with RAG instructions
        """
        return (
            "You are a helpful AI assistant with access to a document database. "
            "When answering questions, use the provided context from the documents "
            "to give accurate, well-sourced responses. If the context doesn't contain "
            "relevant information, acknowledge this and answer based on your general knowledge. "
            "Always cite the document sources when using information from the context."
        )

    @property
    def storage_persist_dir(self) -> str:
        """Get the directory path for persistent RAG storage (legacy).

        This is kept for backward compatibility with the old unified index.

        Returns:
            Absolute path to storage directory
        """
        storage_dir = os.path.join(
            os.path.expanduser(self.path_settings.base_path),
            "rag",
            "storage",
        )
        os.makedirs(storage_dir, exist_ok=True)
        return storage_dir

    def _check_and_download_embedding_model(self, model_path: str) -> bool:
        """Check if embedding model files exist and trigger download if needed.

        Uses llm_file_bootstrap_data.py as the source of truth for required files.

        Args:
            model_path: Path to the embedding model directory

        Returns:
            True if model files exist, False if download was triggered
        """
        from airunner.components.llm.data.bootstrap.llm_file_bootstrap_data import (
            LLM_FILE_BOOTSTRAP_DATA,
        )
        from airunner.enums import SignalCode

        if not hasattr(self, "_embedding_download_pending"):
            self._embedding_download_pending = False
        if not hasattr(self, "_embedding_download_handler_registered"):
            self._embedding_download_handler_registered = False

        # Use llm_file_bootstrap_data as source of truth
        repo_id = "intfloat/e5-large"

        if repo_id not in LLM_FILE_BOOTSTRAP_DATA:
            self.logger.error(
                f"Embedding model {repo_id} not in LLM_FILE_BOOTSTRAP_DATA"
            )
            return False

        # Check which required files are missing or incomplete
        # files is a dict of {filename: expected_size}
        required_files = LLM_FILE_BOOTSTRAP_DATA[repo_id]["files"]
        missing_files = []

        for required_file, expected_size in required_files.items():
            file_path = os.path.join(model_path, required_file)
            if not os.path.exists(file_path):
                missing_files.append(required_file)
            elif expected_size > 0:
                # Check if file is complete by comparing size
                actual_size = os.path.getsize(file_path)
                if actual_size < expected_size:
                    self.logger.warning(
                        f"File {required_file} is incomplete: {actual_size} bytes vs expected {expected_size} bytes"
                    )
                    missing_files.append(required_file)

        if not missing_files:
            # All files exist
            return True

        # Files are missing - trigger download
        self.logger.info(
            f"Missing {len(missing_files)} files for embedding model {repo_id}, triggering download"
        )
        self.logger.debug(f"Missing files: {missing_files}")

        # Emit signal to trigger download dialog
        # Avoid triggering multiple simultaneous downloads
        if not getattr(self, "_embedding_download_pending", False):
            self.emit_signal(
                SignalCode.LLM_MODEL_DOWNLOAD_REQUIRED,
                {
                    "repo_id": repo_id,
                    "model_path": model_path,
                    "missing_files": missing_files,
                    "model_type": "embedding",
                },
            )
            self._embedding_download_pending = True

        # Register handler for download completion to retry embedding initialization
        if (
            hasattr(self, "register")
            and not self._embedding_download_handler_registered
        ):
            self.register(
                SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                self._on_embedding_download_complete,
            )
            self._embedding_download_handler_registered = True

        return False

    def _on_embedding_download_complete(self, data: dict):
        """Handle embedding model download completion and retry initialization.

        Args:
            data: Download completion data
        """
        repo_id = data.get("repo_id")
        if repo_id == "intfloat/e5-large":
            self.logger.info(
                "Embedding model download complete - will initialize on next access"
            )
            # Clear cached embedding so it will be re-initialized on next access
            self._embedding = None
            self._embedding_download_pending = False
            # Unregister this handler
            if hasattr(self, "unregister"):
                from airunner.enums import SignalCode

                self.unregister(
                    SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                    self._on_embedding_download_complete,
                )
                self._embedding_download_handler_registered = False



"""Task complexity detection for automatic harness application.

This module detects when a user's request should be wrapped with the
Long-Running Harness for better coherency and progress tracking.

The harness is automatically applied for:
- Multi-step tasks (e.g., "research 5 papers", "implement these features")
- Complex coding tasks (e.g., "refactor this module", "add tests for X")
- Research projects (e.g., "investigate X and write a report")
- Any task with explicit multiple items or steps

This follows Anthropic's pattern of using a harness to maintain coherency
during complex multi-step tasks within a session.
"""

import re
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple


class TaskType(Enum):
    """Types of tasks that benefit from harness wrapping."""

    SIMPLE = "simple"  # No harness needed
    MULTI_RESEARCH = "multi_research"  # Multiple research topics
    CODING_PROJECT = "coding_project"  # Code implementation/refactoring
    MULTI_STEP = "multi_step"  # Explicit numbered/listed steps
    COMPLEX_ANALYSIS = "complex_analysis"  # Deep analysis tasks


@dataclass
class TaskAnalysis:
    """Result of analyzing a user prompt for task complexity."""

    task_type: TaskType
    should_use_harness: bool
    detected_items: List[str]  # Individual tasks/topics detected
    confidence: float  # 0.0 to 1.0
    reason: str  # Human-readable explanation


# Patterns that indicate multiple items
MULTI_ITEM_PATTERNS = [
    # Explicit numbers: "research 5 papers", "implement 3 features", "write 3 papers"
    r"(\d+)\s+(papers?|topics?|items?|features?|functions?|tests?|files?|endpoints?|components?|modules?|reports?|articles?)",
    # "write X papers/reports" pattern
    r"write\s+(\d+)\s+(?:research\s+)?(?:papers?|reports?|articles?)",
    # Lists with "and": "research X, Y, and Z"
    r"research\s+.+(?:,\s*.+)+(?:,?\s+and\s+.+)",
    # Explicit lists: "the following: 1. X 2. Y"
    r"(?:following|these)[:.]?\s*(?:\d+[.)]\s*.+){2,}",
    # Bullet points
    r"[-•]\s*.+(?:\n[-•]\s*.+)+",
]

# Keywords indicating coding projects
CODING_PROJECT_KEYWORDS = [
    "implement",
    "refactor",
    "build",
    "create",
    "develop",
    "add feature",
    "add test",
    "write test",
    "unit test",
    "fix bug",
    "debug",
    "optimize",
    "migrate",
    "upgrade",
    "port",
    "convert",
    "integrate",
]

# Compound patterns that indicate coding projects (require fewer matches)
CODING_COMPOUND_PATTERNS = [
    r"refactor.*and.*(?:add|write|create)",  # "refactor X and add Y"
    r"(?:add|write|create).*(?:test|feature).*and",  # "add tests and..."
    r"implement.*(?:with|including).*test",  # "implement X with tests"
]

# Keywords indicating multi-step tasks
MULTI_STEP_KEYWORDS = [
    "step by step",
    "steps",
    "first.*then",
    "after that",
    "following steps",
    "process",
    "workflow",
    "pipeline",
]

# Keywords indicating complex analysis
ANALYSIS_KEYWORDS = [
    "analyze",
    "investigate",
    "compare",
    "evaluate",
    "assess",
    "review",
    "audit",
    "examine",
    "comprehensive",
    "thorough",
    "in-depth",
    "detailed analysis",
]

# Research-specific patterns
RESEARCH_PATTERNS = [
    r"research\s+(\d+)",  # "research 5 papers"
    r"write\s+(\d+)\s+(?:papers?|reports?|articles?)",
    r"investigate\s+(?:multiple|several|\d+)",
    r"compare\s+(?:multiple|several|\d+)",
]


def analyze_task(prompt: str) -> TaskAnalysis:
    """Analyze a prompt to determine if it needs harness wrapping.

    Args:
        prompt: User's input text

    Returns:
        TaskAnalysis with detection results
    """
    prompt_lower = prompt.lower()
    detected_items = []
    confidence = 0.0
    reasons = []

    # Check for explicit multiple items (highest confidence)
    for pattern in MULTI_ITEM_PATTERNS:
        match = re.search(pattern, prompt_lower)
        if match:
            confidence = max(confidence, 0.9)
            reasons.append(f"Explicit multi-item pattern: {match.group(0)[:50]}")
            # Try to extract count
            count_match = re.search(r"(\d+)", match.group(0))
            if count_match:
                count = int(count_match.group(1))
                detected_items = [f"item_{i+1}" for i in range(count)]

    # Check for research patterns
    for pattern in RESEARCH_PATTERNS:
        match = re.search(pattern, prompt_lower)
        if match:
            confidence = max(confidence, 0.85)
            reasons.append(f"Research pattern detected: {match.group(0)}")
            if match.groups():
                try:
                    count = int(match.group(1))
                    detected_items = [f"research_topic_{i+1}" for i in range(count)]
                except (ValueError, IndexError):
                    pass

    # Check for coding project keywords
    coding_score = sum(
        1 for kw in CODING_PROJECT_KEYWORDS if kw in prompt_lower
    )
    
    # Check for compound coding patterns (these are strong signals)
    for pattern in CODING_COMPOUND_PATTERNS:
        if re.search(pattern, prompt_lower):
            confidence = max(confidence, 0.75)
            reasons.append(f"Compound coding pattern: {pattern[:30]}")
            coding_score += 2  # Boost the score
            break
    
    if coding_score >= 2:
        confidence = max(confidence, 0.7)
        reasons.append(f"Multiple coding keywords ({coding_score})")
    elif coding_score == 1:
        confidence = max(confidence, 0.4)
        reasons.append("Single coding keyword")

    # Check for multi-step keywords
    for kw in MULTI_STEP_KEYWORDS:
        if re.search(kw, prompt_lower):
            confidence = max(confidence, 0.6)
            reasons.append(f"Multi-step keyword: {kw}")
            break

    # Check for analysis keywords
    analysis_score = sum(1 for kw in ANALYSIS_KEYWORDS if kw in prompt_lower)
    if analysis_score >= 2:
        confidence = max(confidence, 0.65)
        reasons.append(f"Complex analysis ({analysis_score} keywords)")

    # Check for comma-separated lists (e.g., "research X, Y, Z")
    comma_items = _extract_comma_list(prompt)
    if len(comma_items) >= 3:
        confidence = max(confidence, 0.75)
        reasons.append(f"Comma-separated list: {len(comma_items)} items")
        detected_items = comma_items

    # Determine task type
    task_type = _determine_task_type(prompt_lower, confidence, reasons)

    # Decide if harness should be used
    should_use_harness = confidence >= 0.6 and task_type != TaskType.SIMPLE

    return TaskAnalysis(
        task_type=task_type,
        should_use_harness=should_use_harness,
        detected_items=detected_items,
        confidence=confidence,
        reason="; ".join(reasons) if reasons else "Simple task, no harness needed",
    )


def _extract_comma_list(prompt: str) -> List[str]:
    """Extract items from a comma-separated list in the prompt.

    Args:
        prompt: User's input text

    Returns:
        List of extracted items
    """
    # Look for patterns like "research X, Y, and Z" or "implement A, B, C"
    patterns = [
        r"(?:research|implement|analyze|compare|write about|investigate)\s+(.+)",
    ]

    for pattern in patterns:
        match = re.search(pattern, prompt, re.IGNORECASE)
        if match:
            items_str = match.group(1)
            # First, handle "and" separately to avoid "and X" in results
            # Replace ", and " with just ","
            items_str = re.sub(r",\s*and\s+", ", ", items_str)
            # Replace standalone " and " with ","
            items_str = re.sub(r"\s+and\s+", ", ", items_str)
            # Now split by comma
            items = re.split(r",\s*", items_str)
            # Clean up items
            items = [item.strip() for item in items if item.strip()]
            # Filter out items that are too short or look like noise
            items = [
                item
                for item in items
                if len(item) > 2 and item.lower() not in ["the", "a", "an"]
            ]
            if len(items) >= 2:
                return items[:10]  # Cap at 10 items

    return []


def _determine_task_type(
    prompt_lower: str, confidence: float, reasons: List[str]
) -> TaskType:
    """Determine the specific type of task.

    Args:
        prompt_lower: Lowercased prompt
        confidence: Current confidence score
        reasons: List of detection reasons

    Returns:
        TaskType enum value
    """
    if confidence < 0.5:
        return TaskType.SIMPLE

    # Check specific types
    if any("research" in r.lower() for r in reasons):
        return TaskType.MULTI_RESEARCH

    if any("coding" in r.lower() for r in reasons):
        return TaskType.CODING_PROJECT

    if any("multi-step" in r.lower() for r in reasons):
        return TaskType.MULTI_STEP

    if any("analysis" in r.lower() for r in reasons):
        return TaskType.COMPLEX_ANALYSIS

    # Default based on keywords in prompt
    if "research" in prompt_lower or "paper" in prompt_lower:
        return TaskType.MULTI_RESEARCH
    if any(kw in prompt_lower for kw in CODING_PROJECT_KEYWORDS[:5]):
        return TaskType.CODING_PROJECT

    return TaskType.MULTI_STEP


def should_use_harness(prompt: str) -> Tuple[bool, Optional[TaskAnalysis]]:
    """Quick check if a prompt should use the harness.

    Args:
        prompt: User's input text

    Returns:
        Tuple of (should_use, analysis) - analysis is None if not using harness
    """
    analysis = analyze_task(prompt)
    if analysis.should_use_harness:
        return True, analysis
    return False, None
